package higa.dbawt;

import java.awt.TextArea;

import java.awt.event.TextEvent;
import java.awt.event.TextListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import java.text.Format;

import higa.sql.dataset.Column;
import higa.sql.dataset.DataChangeEvent;
import higa.sql.dataset.DataChangeListener;
import higa.sql.dataset.NavigationEvent;
import higa.sql.dataset.NavigationListener;
import higa.sql.dataset.IDataSet;
import higa.sql.dataset.ColumnAware;
import higa.sql.dataset.Status;

import higa.util.Formatter;

public class HTextArea
  extends TextArea
  implements ColumnAware,
  DataChangeListener, NavigationListener,
  FocusListener, KeyListener {

  private IDataSet dataSet;
  private String columnName;
  private Format formatter;
  private String formatPattern;

  private String current;
  private boolean addNotifyCalled = false;

  public HTextArea() {
    addFocusListener(this);
    addKeyListener(this);
  }

  public void setDataSet(IDataSet ds) {
    if (dataSet != null) {
      dataSet.removeDataChangeListener(this);
      dataSet.removeNavigationListener(this);
    }

    dataSet = ds;
    if (addNotifyCalled &&
      dataSet != null) {

      openDataSet();
    }

    if (dataSet != null) {
      dataSet.addDataChangeListener(this);
      dataSet.addNavigationListener(this);
    }
  }

  public IDataSet getDataSet() {
    return dataSet;
  }

  public void setColumnName(String s) {
    columnName = s;
    if (addNotifyCalled &&
      dataSet != null) {

      openDataSet();
    }
  }

  public String getColumnName() {
    return columnName;
  }

  public void setFormatter(Format formatter) {
    this.formatter = formatter;
    formatPattern = null;
    updateText();
  }

  public Format getFormatter() {
    return formatter;
  }

  public void setFormatPattern(String pattern) {
    formatPattern =
      pattern != null && pattern.length() != 0 ?
      pattern : null;

    formatter = null;
    updateText();
  }

  public String getFormatPattern() {
    return formatPattern;
  }

  public void addNotify() {
    super.addNotify();
    if (!addNotifyCalled) {
      addNotifyCalled = true;
      if (dataSet != null) {
        openDataSet();
      }
    }
  }

  public void dataChanged(DataChangeEvent e) {
    updateText();
  }

  public void navigated(NavigationEvent e) {
    updateText();
  }

  public void focusGained(FocusEvent evt) {}
  
  public void focusLost(FocusEvent e) {
    String s = getText();
    if (!s.equals(current)) {
      postText();
    }
  }

  public void keyTyped(KeyEvent e) {}

  public void keyPressed(KeyEvent e) {
    if (KeyEvent.VK_ENTER == e.getKeyCode() &&
      e.isControlDown()) {

      postText();
      e.consume();
    }
  }

  public void keyReleased(KeyEvent e) {}

  protected void updateText() {
    if (dataSet == null) {
      return;
    }

    Column column = dataSet.getColumn(columnName);
    if (column == null) {
      return;
    }

    if (formatter == null &&
      formatPattern != null) {

      int type = column.getType();
      formatter = Formatter.getDefaultFormatter(
        type, formatPattern);
    }

    String s = "";
    if (formatter == null) {
      s = dataSet.getString(columnName);
    } else {
      Object value = dataSet.get(columnName);
      s = Formatter.format(formatter, value);
    }

    setText(s);
    current = s;
  }

  protected void postText() {
    if (dataSet == null ||
      dataSet.getColumn(columnName) == null) {

      return;
    }

    String s = getText();
    Object value = formatter == null ?
      s : Formatter.parseObject(formatter, s);

    try {
      dataSet.set(columnName, value);
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  private void openDataSet() {
    if (dataSet == null || !addNotifyCalled) {
      return;
    }

    if (!dataSet.isOpen()) {
      try {
        dataSet.open();
      } catch (Exception ex) {
        ex.printStackTrace();
        return;
      }
    }

    if (dataSet.isOpen()) {
      updateText();
    }
  }
}
