package higa.dbswing;

import com.sun.java.swing.JList;
import com.sun.java.swing.ListSelectionModel;

import com.sun.java.swing.event.ListSelectionEvent;
import com.sun.java.swing.event.ListSelectionListener;

import java.awt.Dimension;

import java.sql.Statement;
import java.sql.ResultSet;

import java.io.Serializable;

import higa.sql.dataset.PickListAware;
import higa.sql.dataset.ColumnAware;
import higa.sql.dataset.IDataSet;
import higa.sql.dataset.HDatabase;
import higa.sql.dataset.Column;
import higa.sql.dataset.DataChangeEvent;
import higa.sql.dataset.DataChangeListener;
import higa.sql.dataset.NavigationEvent;
import higa.sql.dataset.NavigationListener;

public class HJList extends JList
  implements PickListAware, ColumnAware,
  DataChangeListener, NavigationListener,
  ListSelectionListener, Serializable {

  private String pickListTableName;
  private String pickListColumnName;
  private String pickListDisplayColumnName;
  private String columnName;
  private IDataSet dataSet;

  private transient boolean addNotifyCalled = false;
  private transient boolean selfEvent = false;

  public HJList() {
  }

  public void setPickListTableName(String s) {
    if (s != null && s.equals("")) {
      s = null;
    }

    pickListTableName = s;
    if (addNotifyCalled &&
      dataSet != null) {

      openDataSet();
    }
  }

  public String getPickListTableName() {
    return pickListTableName;
  }

  public void setPickListColumnName(String s) {
    if (s != null && s.equals("")) {
      s = null;
    }

    pickListColumnName = s;
    if (addNotifyCalled &&
      dataSet != null) {

      openDataSet();
    }
  }

  public String getPickListColumnName() {
    return pickListColumnName;
  }

  public void setPickListDisplayColumnName(String s) {
    if (s != null && s.equals("")) {
      s = null;
    }

    pickListDisplayColumnName = s;
    if (addNotifyCalled &&
      dataSet != null) {

      openDataSet();
    }
  }

  public String getPickListDisplayColumnName() {
    return pickListDisplayColumnName;
  }

  public void setColumnName(String s) {
    columnName = s;
    if (addNotifyCalled &&
      dataSet != null) {

      openDataSet();
    }
  }

  public String getColumnName() {
    return columnName;
  }

  public void setDataSet(IDataSet ds) {
    if (dataSet != null) {
      removeListSelectionListener(this);
      dataSet.removeDataChangeListener(this);
      dataSet.removeNavigationListener(this);
    }

    dataSet = ds;
    if (addNotifyCalled &&
      dataSet != null) {

      openDataSet();
    }
  }

  public IDataSet getDataSet() {
    return dataSet;
  }

  public void addNotify() {
    super.addNotify();
    if (!addNotifyCalled) {
      addNotifyCalled = true;
      if (dataSet != null) {
        openDataSet();
      }
    }
  }

  public Dimension getPreferredSize() {
    Dimension ps = super.getPreferredSize();
    if (ps == null || ps.width < 40) {
      ps = ps != null ? ps : new Dimension(100, 25);
      ps.width = ps.width > 100 ? ps.width : 100;
    }
    
    return ps;
  }

  public void valueChanged(ListSelectionEvent e) {
    if (e.getValueIsAdjusting() || selfEvent) {
      return;
    }

    postSelection();
  }

  public void dataChanged(DataChangeEvent e) {
    updateSelection();
  }

  public void navigated(NavigationEvent e) {
    updateSelection();
  }

  protected void updateSelection() {
    if (dataSet == null ||
      dataSet.getColumn(columnName) == null) {

      return;
    }

    String value = dataSet.getString(columnName);
    HListModel model = (HListModel)getModel();
    int index = model.getValueIndex(value);
    selfEvent = true;
    setSelectedIndex(index);
    selfEvent = false;
  }

  protected void postSelection() {
    int index = getSelectedIndex();
    if(index < 0 || dataSet == null ||
      dataSet.getColumn(columnName) == null) {

      return;
    }

    try {
      HListModel model = (HListModel)getModel();
      Object value = model.getValueAt(index);
      dataSet.set(columnName, value);
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  private void openDataSet() {
    if (dataSet == null || !addNotifyCalled) {
      return;
    }

    if (!dataSet.isOpen()) {
      try {
        dataSet.open();
      } catch (Exception ex) {
        ex.printStackTrace();
        return;
      }
    }

    if (dataSet.isOpen()) {
      HListModel model = new HListModel(dataSet, columnName,
        pickListTableName, pickListColumnName,
        pickListDisplayColumnName);

      setModel(model);
      updateSelection();
      addListSelectionListener(this);
      setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      dataSet.addDataChangeListener(this);
      dataSet.addNavigationListener(this);
    }
  }
}