package higa.jhcl.control;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import higa.util.Resources;

import higa.sql.dataset.DataSetAware;
import higa.sql.dataset.IDataSet;
import higa.sql.dataset.StatusEvent;
import higa.sql.dataset.StatusListener;

public class HNavigator
  extends HButtonBar
  implements DataSetAware, ActionListener,
  StatusListener {

  public static final int FIRST_BUTTON = 0;
  public static final int PRIOR_BUTTON = 1;
  public static final int NEXT_BUTTON = 2;
  public static final int LAST_BUTTON = 3;
  public static final int INSERT_BUTTON = 4;
  public static final int DELETE_BUTTON = 5;
  public static final int POST_BUTTON = 6;
  public static final int CANCEL_BUTTON = 7;
  public static final int COMMIT_BUTTON = 8;
  public static final int ROLLBACK_BUTTON = 9;

  private static final Resources resources =
    new Resources(
    "higa.jhcl.control.resources.control");

  public static final String FIRST =
    resources.getString("first");

  public static final String PRIOR =
    resources.getString("prior");

  public static final String NEXT =
    resources.getString("next");

  public static final String LAST =
    resources.getString("last");

  public static final String INSERT =
    resources.getString("insert");

  public static final String DELETE =
    resources.getString("delete");

  public static final String POST =
    resources.getString("post");

  public static final String CANCEL =
    resources.getString("cancel");

  public static final String COMMIT =
    resources.getString("commit");

  public static final String ROLLBACK =
    resources.getString("rollback");

  private static final String[] ACTIONS =
    {FIRST, PRIOR, NEXT, LAST, INSERT,
    DELETE, POST, CANCEL, COMMIT, ROLLBACK};

  private static final int BUTTON_COUNT =
    ACTIONS.length;

  private IDataSet dataSet;

  public HNavigator() {
    setLabels(ACTIONS);
    setRows(2);
    addActionListener(this);
  }

  public void setDataSet(IDataSet ds) {
    if (dataSet != null) {
      dataSet.removeStatusListener(this);
    }

    dataSet = ds;

    if (dataSet != null) {
      dataSet.addStatusListener(this);
    }
  }

  public IDataSet getDataSet() {
    return dataSet;
  }

  public void actionPerformed(ActionEvent evt) {
    if (null == dataSet) {
      return;
    }

    String label = evt.getActionCommand();
    int index = getLabelIndex(label);

    try {
      switch (index) {
        case 0:
          dataSet.first();
          break;
        case 1:
          dataSet.prior();
          break;
        case 2:
          dataSet.next();
          break;
        case 3:
          dataSet.last();
          break;
        case 4:
          dataSet.insertRow();
          break;
        case 5:
          dataSet.deleteRow();
          break;
        case 6:
          dataSet.post();
          break;
        case 7:
          dataSet.cancel();
          break;
        case 8:
          dataSet.commit();
          break;
        case 9:
          dataSet.rollback();
          break;
        default:
          throw new IllegalStateException();
      }
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  public void statusMessage(StatusEvent e) {
    for (int i = 0; i < BUTTON_COUNT; i++) {
      setButtonEnabled(i, false);
    }

    setButtonEnabled(INSERT_BUTTON, true);

    if (!dataSet.isFirst()) {
      setButtonEnabled(FIRST_BUTTON, true);
      setButtonEnabled(PRIOR_BUTTON, true);
    }

    if (!dataSet.isLast()) {
      setButtonEnabled(NEXT_BUTTON, true);
      setButtonEnabled(LAST_BUTTON, true);
    }

    if (dataSet.isEditing()) {
      setButtonEnabled(POST_BUTTON, true);
      setButtonEnabled(CANCEL_BUTTON, true);
    }

    if (dataSet.isTransaction()) {
      setButtonEnabled(COMMIT_BUTTON, true);
      setButtonEnabled(ROLLBACK_BUTTON, true);
    }

    if (!dataSet.isEmpty()) {
      setButtonEnabled(DELETE_BUTTON, true);
    }
  }
}