package higa.jhcl.model;

import java.io.Serializable;

import higa.util.EventMulticaster;

public class BasicMatrixSelection extends BasicSelection
  implements MatrixSelection, Serializable {

  protected EventMulticaster selectionCaster =
    new EventMulticaster();

  protected boolean eventEnabled = true;

  public BasicMatrixSelection() {}

  public BasicMatrixSelection(MatrixLocation[] newSet) {
    set(newSet);
  }

  public boolean contains(int rowIndex, int columnIndex) {
    return contains(
      new MatrixLocation(rowIndex, columnIndex));
  }

  public boolean contains(MatrixLocation location) {
    return super.contains(location);
  }

  public MatrixLocation[] getAll() {
    int count = getCount();
    MatrixLocation[] contents = new MatrixLocation[count];

    for (int i = 0; i < count; i++) {
      contents[i] = (MatrixLocation)doGet(i);
    }

    return contents;
  }

  public void addSelectionListener(
    MatrixSelectionListener listener) {

    selectionCaster.add(listener);
  }

  public void removeSelectionListener(
    MatrixSelectionListener listener) {

    selectionCaster.remove(listener);
  }

  public void set(MatrixLocation[] locations) {
    doRemoveAll();
    for (int i = 0 ; i < locations.length ; i++) {
      doAdd(locations[i]);
    }

    processSelectionEvent(new MatrixSelectionEvent(
      this, SelectionEvent.SELECTION_CHANGED));
  }

  public void add(int rowIndex, int columnIndex) {
    MatrixLocation ml =
      new MatrixLocation(rowIndex, columnIndex);

    add(ml);
  }

  public void add(MatrixLocation location) {
    if (doAdd(location)) {
      processSelectionEvent(new MatrixSelectionEvent(
        this, SelectionEvent.ITEM_ADDED, location));
    }
  }

  public void add(MatrixLocation[] locations) {
    boolean changed = false;
    for (int i = 0 ; i < locations.length ; i++) {
      changed |= doAdd(locations[i]);
    }

    if (changed) {
      processSelectionEvent(new MatrixSelectionEvent(
        this, SelectionEvent.SELECTION_CHANGED));
    }
  }

  public void addRange(MatrixLocation begin, MatrixLocation end) {
    addRange(begin.rowIndex, begin.columnIndex,
      end.rowIndex, end.columnIndex);
  }

  public void addRange(int beginRowIndex, int beginColumnIndex,
    int endRowIndex, int endColumnIndex) {

    int loRowIndex = Math.min(beginRowIndex, endRowIndex);
    int hiRowIndex = Math.max(beginRowIndex, endRowIndex);
    int loColIndex = Math.min(beginColumnIndex, endColumnIndex);
    int hiColIndex = Math.max(beginColumnIndex, endColumnIndex);

    boolean changed = false;
    for (int r = loRowIndex ; r <= hiRowIndex; r++) {

      for (int c = loColIndex; c <= hiColIndex; c++) {
        changed |= doAdd(
          new MatrixLocation(r, c));
      }
    }

    if (changed) {
      processSelectionEvent(new MatrixSelectionEvent(
        this, SelectionEvent.RANGE_ADDED,
        new MatrixLocation(loRowIndex, loColIndex),
        new MatrixLocation(hiRowIndex, hiColIndex)));
    }
  }

  public void remove(int rowIndex, int columnIndex) {
    MatrixLocation ml =
      new MatrixLocation(rowIndex, columnIndex);

    remove(ml);
  }

  public void remove(MatrixLocation location) {
    if (doRemove(location)) {
      processSelectionEvent(new MatrixSelectionEvent(
        this, SelectionEvent.ITEM_REMOVED, location));
    }
  }

  public void remove(MatrixLocation[] locations) {
    boolean changed = false;
    for (int i = 0 ; i < locations.length ; i++ ) {
      changed |= doRemove(locations[i]);
    }

    if (changed) {
      processSelectionEvent(new MatrixSelectionEvent(
        this, SelectionEvent.SELECTION_CHANGED));
    }
  }

  public void removeRange(MatrixLocation begin, MatrixLocation end) {
    removeRange(begin.rowIndex, begin.columnIndex,
      end.rowIndex, end.columnIndex);
  }

  public void removeRange(int beginRowIndex, int beginColumnIndex,
    int endRowIndex, int endColumnIndex) {
    int loRowIndex = Math.min(beginRowIndex, endRowIndex);
    int hiRowIndex = Math.max(beginRowIndex, endRowIndex);
    int loColIndex = Math.min(beginColumnIndex, endColumnIndex);
    int hiColIndex = Math.max(beginColumnIndex, endColumnIndex);

    boolean changed = false;
    for (int r = loRowIndex; r <= hiRowIndex; r++) {
      for (int c = loColIndex; c <= hiColIndex; c++) {
        changed |= doRemove(new MatrixLocation(r, c));
      }
    }

    if (changed) {
      processSelectionEvent(new MatrixSelectionEvent(
        this, SelectionEvent.RANGE_REMOVED,
        new MatrixLocation(loRowIndex, loColIndex),
        new MatrixLocation(hiRowIndex, hiColIndex)));
    }
  }

  public void removeAll() {
    if (doRemoveAll()) {
      processSelectionEvent(new MatrixSelectionEvent(
        this, SelectionEvent.SELECTION_CLEARED));
    }
  }

  public void setEventEnabled(boolean enabled) {
    eventEnabled = enabled;
  }

  protected void processSelectionEvent(MatrixSelectionEvent e) {
    if (eventEnabled && selectionCaster.hasListeners()) {
      selectionCaster.dispatch(e);
    }
  }
}