package higa.jhcl.view;

import java.io.Serializable;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;

import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;

import higa.jhcl.model.ItemPainter;
import higa.jhcl.model.ItemPaintSite;
import higa.jhcl.model.SingletonModel;
import higa.jhcl.model.SingletonModelEvent;
import higa.jhcl.model.SingletonModelListener;
import higa.jhcl.model.SingletonModelMulticaster;

import higa.util.Alignment;

public class ButtonView
  extends HComponent
  implements SingletonView, SingletonModelListener,
  Serializable {

  private SingletonModel model;
  private ItemPainter painter;
  private String actionCommand;
  private boolean flat;
  private ButtonItemPainter borderPainter =
    new ButtonItemPainter();

  private SingletonModelMulticaster modelCaster =
    new SingletonModelMulticaster();

  public ButtonView() {}

  public void setLabel(String label) {
    if (!isReadOnly()) {
      model.set(label);
      setActionCommand(label);
      repaint(100);
    }
  }

  public String getLabel() {
    Object contents = model != null ?
      model.get() : null;

    return contents != null ?
      contents.toString() : "";
  }

  public void setModel(SingletonModel sm) {
    if (model != null) {
      model.removeModelListener(this);
      model.removeModelListener(modelCaster);
    }

    model = sm;
    if (model != null) {
      model.addModelListener(this);
      model.addModelListener(modelCaster);
    }

    repaint(100);
  }

  public SingletonModel getModel() {
    return model;
  }

  public void modelContentChanged(SingletonModelEvent e) {
    repaint(100);
  }

  public void setActionCommand(String command) {
    actionCommand = command;
  }

  public String getActionCommand() {
    return actionCommand;
  }

  public void setPainter(ItemPainter painter) {
    this.painter = painter;
  }

  public ItemPainter getPainter() {
    return painter;
  }

  public void setFlat(boolean flat) {
    if (this.flat != flat) {
      this.flat = flat;
      repaint();
    }
  }

  public boolean isFlat() {
    return flat;
  }

  public Dimension getPreferredSize() {
    int state = getState();
    Graphics g = getGraphics();
    Dimension bs =
      borderPainter.getPreferredSize(
      null, g, state, this);

    if (model != null) {
      Object value = model.get();
      if (value != null &&
        painter != null) {

        Dimension is =
          painter.getPreferredSize(
          value, g, state, this);

        return new Dimension(
          is.width + bs.width,
          is.height + bs.height);
      }
    }

    return bs;
  }

  public void paint(Graphics g) {
    int state = getState();
    super.paint(g);
    Dimension outerSize = getSize();
    Dimension borderSize =
      borderPainter.getPreferredSize(
      null, g, state, this);

    Dimension innerSize =
      new Dimension(
      outerSize.width - borderSize.width,
      outerSize.height - borderSize.height);

    if (isEnabled() &&
      (!flat || (state & ItemPainter.ROLLOVER) != 0)) {
      borderPainter.paint(null, g,
        new Rectangle(0, 0,
        outerSize.width, outerSize.height),
        state, this);
    }

    boolean pressed =
      (state & ItemPainter.SELECTED) != 0;

    int left = borderSize.width / 2 +
      (pressed ? 1 : 0);

    int top  = borderSize.height / 2 +
      (pressed ? 1 : 0);

    Object value = model != null ?
      model.get() : null;

    if (painter != null) {
      Rectangle r = new Rectangle(
        left, top, innerSize.width,
        innerSize.height);

      g.clipRect(r.x, r.y, r.width, r.height);
      g.setColor(getBackground());
      g.setFont(getFont());
      painter.paint(value, g, r,
        state, this);
    }
  }

  public void addModelListener(SingletonModelListener listener) {
    modelCaster.add(listener);
  }

  public void removeModelListener(SingletonModelListener listener) {
    modelCaster.remove(listener);
  }

  protected void processKeyPressed(KeyEvent e) {
    super.processKeyPressed(e);
    if (!isEnabled()) {
      return;
    }

    switch (e.getKeyCode()) {
      case KeyEvent.VK_SPACE:
        repaint();
        break;
    }
  }

  protected void processKeyReleased(KeyEvent e) {
    super.processKeyReleased(e);
    if (e.getKeyCode() == KeyEvent.VK_SPACE) {
      repaint();
      if (isEnabled()) {
        processActionEvent(new ActionEvent(
          this, ActionEvent.ACTION_PERFORMED,
          getActionCommand()));
      }
    }
  }

  protected void processMousePressed(MouseEvent e) {
    super.processMousePressed(e);
    if (!isEnabled()) {
      return;
    }

    repaint();
  }

  protected void processMouseReleased(MouseEvent e) {
    boolean action =
      (getState() & ItemPainter.SELECTED) != 0;

    super.processMouseReleased(e);

    repaint();
    if (action && isEnabled()) {
      processActionEvent(new ActionEvent(
        this, ActionEvent.ACTION_PERFORMED,
        getActionCommand()));

      processMouseEntered(e);
    }
  }

  protected void processMouseEntered(MouseEvent e) {
    super.processMouseEntered(e);
    if (!isEnabled()) {
      return;
    }

    repaint();
  }

  protected void processMouseExited(MouseEvent e) {
    super.processMouseExited(e);
    repaint();
  }
}