package higa.jhcl.view;

import java.io.Serializable;

import java.awt.Point;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Canvas;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.SystemColor;
import java.awt.Component;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;

import java.util.Vector;

import higa.jhcl.model.MatrixLocation;
import higa.jhcl.model.ItemPainter;
import higa.jhcl.model.ItemEditor;
import higa.jhcl.model.ItemEditSite;
import higa.jhcl.model.VectorModelEvent;
import higa.jhcl.model.VectorModelListener;
import higa.jhcl.model.MatrixModel;
import higa.jhcl.model.MatrixModelEvent;
import higa.jhcl.model.MatrixModelListener;
import higa.jhcl.model.BasicMatrixContainer;
import higa.jhcl.model.SelectionEvent;
import higa.jhcl.model.MatrixSelection;
import higa.jhcl.model.MatrixSelectionEvent;
import higa.jhcl.model.MatrixSelectionListener;
import higa.jhcl.model.BasicMatrixSelection;
import higa.jhcl.model.MatrixModelMulticaster;
import higa.jhcl.model.MatrixSelectionMulticaster;
import higa.jhcl.model.SubfocusEvent;
import higa.jhcl.model.MatrixSubfocusEvent;
import higa.jhcl.model.MatrixSubfocusListener;

import higa.util.EventMulticaster;
import higa.util.KeyMulticaster;
import higa.util.ActionMulticaster;
import higa.util.SelectFlags;
import higa.util.Alignment;

public class GridCore extends HComponent
  implements MatrixView, MatrixModelListener,
  MatrixSelectionListener, ItemEditSite,
  KeyListener, FocusListener, Serializable {

  private Scroller scroller;
  private Color lineColor =
    SystemColor.control;

  private transient Image canvas;
  private MatrixLocation origin;

  private int defaultRowHeight = 20;

  private ItemEditor defaultEditor;
  private ItemPainter defaultPainter;

  private ColumnViewManager columnViewManager =
    new ColumnViewManager();

  private RowViewManager rowViewManager =
    new RowViewManager(this);

  private MatrixModel model;
  private transient MatrixSelection selection;

  private MatrixModelMulticaster modelCaster =
    new MatrixModelMulticaster();

  private MatrixSelectionMulticaster selectionCaster =
    new MatrixSelectionMulticaster();

  private EventMulticaster subfocusCaster =
    new EventMulticaster();

  private KeyMulticaster keyCaster =
    new KeyMulticaster();

  private ActionMulticaster actionCaster =
    new ActionMulticaster();

  private boolean batchMode = false;
  private boolean hasFocus = false;
  private boolean rangeSelecting = false;
  private boolean directDraw = true;
  private transient ItemEditor editor;
  private MatrixLocation subfocus;
  private MatrixLocation[] oldSelected =
    new MatrixLocation[0];

  private MatrixLocation selectAnchor;
  private MatrixLocation editorLocation;
  private MatrixLocation mouseDown;
  private Point editClickPoint;

  private MatrixLocation resize;
  private MatrixLocation resize1;

  private Canvas divider = new Canvas();

  private int dividerWidth = 1;
  private int dividerHeight = 1;
  private int dividerDelta;

  public GridCore() {
    setBackground(SystemColor.window);
    divider.setBackground(SystemColor.controlShadow);
    setLayout(null);
    divider.setVisible(false);
    add(divider);
    super.addKeyListener(keyCaster);
  }

  public GridCore(Scroller scroller) {
    this();
    this.scroller = scroller;
  }

  public void setModel(MatrixModel mm) {
    if (editor != null) {
      safeEndEdit();
    }

    if (model != null) {
      model.removeModelListener(this);
      model.removeModelListener(modelCaster);
    }

    model = mm;
    if (model != null) {
      model.addModelListener(this);
      model.addModelListener(modelCaster);
    }

    if (model != null) {
      invalidate();
      reset();
    }
  }

  public MatrixModel getModel() {
    return model;
  }

  public void setPainter(ItemPainter painter) {
    defaultPainter = painter;
    repaintCells();
  }

  public ItemPainter getPainter() {
    return defaultPainter;
  }

  protected ItemPainter getPainter(int columnIndex) {
    ItemPainter ip = columnViewManager.getPainter(columnIndex);
    return ip != null ? ip : defaultPainter;
  }

  public void setEditor(ItemEditor editor) {
    defaultEditor = editor;
  }

  public ItemEditor getEditor() {
    return defaultEditor;
  }

  protected ItemEditor getEditor(int columnIndex) {
    ItemEditor ie = columnViewManager.getEditor(columnIndex);
    return ie != null ? ie : defaultEditor;
  }

  public void setSelection(MatrixSelection ms) {
    if (selection != null) {
      selection.removeSelectionListener(this);
      selection.removeSelectionListener(selectionCaster);
    }

    selection = ms;
    if (selection != null) {
      selection.addSelectionListener(this);
      selection.addSelectionListener(selectionCaster);
    }

    oldSelected = selection.getAll();
    repaintCells();
  }

  public MatrixSelection getSelection() {
    return selection;
  }

  public void repaintCell(MatrixLocation cell) {
    if (batchMode || cell == null) {
      return;
    }

    if (cell.rowIndex < getRowCount() &&
      cell.columnIndex < getColumnCount()) {

      Rectangle rect = getCellRect(cell);
      paint(rect);
    }
  }

  public void repaintSelection() {
    setDirectDraw(false);
    for (int i = 0 ; i < oldSelected.length ; i++) {
      repaintCell(oldSelected[i]);
    }

    setDirectDraw(true);
  }

  private void paint(Rectangle rect) {
    if (rect == null) {
      return;
    }

    if (directDraw) {
      Graphics g = getGraphics();
      if (g != null) {
        g.setClip(rect.x, rect.y,
          rect.width, rect.height);

        paint(g);
      }
    } else {
      repaint(0, rect.x, rect.y,
        rect.width, rect.height);
    }
  }

  public void repaintCells() {
    if (batchMode) {
      return;
    }

    repaint();
  }

  public void repaintCells(MatrixLocation s, MatrixLocation e) {
    if (batchMode) {
      return;
    }

    Rectangle rect = getCellRangeRect(s, e);
    paint(rect);
  }

  public void paint(Graphics pg) {
    if (batchMode) {
      return;
    }

    Rectangle clip = getClipBounds();

    if (clip == null) {
      return;
    }

    Dimension vp = getSize();

    if (canvas == null ||
      canvas.getWidth(null) != vp.width ||
      canvas.getHeight(null) != vp.height) {
      canvas = createImage(vp.width, vp.height);
    }

    Graphics g = canvas.getGraphics();
    if (g == null) {
      return;
    }

    g.setColor(getBackground());
    g.fillRect(0, 0, vp.width, vp.height);
    g.translate(-clip.x, -clip.y);
    g.setClip(clip.x, clip.y, clip.width, clip.height);

    super.paint(g);

    int lastX = clip.x;
    int lastY = clip.y;
    MatrixLocation hit = hitTest(lastX, lastY);

    if (hit != null) {
      g.setFont(getFont());

      int lci = getLastColumnIndex();
      int lri = getLastRowIndex();

      Rectangle cr = getCellRect(hit);
      Rectangle cr2 = new Rectangle(cr);

      int xStart = cr.x;
      for (int r = hit.rowIndex; r <= lri; r++) {
        cr.x = xStart;
        cr.height = rowViewManager.getSize(r);
        if (cr.height > 0) {
          for (int c = hit.columnIndex; c <= lci; c++) {
            cr.width = columnViewManager.getSize(c);
            if (cr.width > 0) {
              int state = getState(r, c);
              Object data = model.get(r, c);
              ItemPainter painter =
                getPainter(c);

              ColumnView cv = columnViewManager.getColumnView(c);
              g.setColor(getBackground());
              Rectangle rect = new Rectangle(cr);
              rect.width--;
              rect.height--;
              g.clipRect(rect.x, rect.y,
                rect.width, rect.height);

              painter.paint(data, g, rect, state, cv);
              g.setClip(clip.x, clip.y,
                clip.width, clip.height);

              cr.x += cr.width;
            }
          }

          cr.y += cr.height;
        }
      }

      lastX = cr.x;
      lastY = cr.y;

      g.setColor(isEnabled() ? lineColor :
        lineColor.brighter());

      cr2.x--;
      cr2.y--;
      lastX--;
      lastY--;

      Dimension d = new Dimension(cr2.x, cr2.y);
      g.drawLine(d.width, d.height, d.width, lastY);
      for (int c = hit.columnIndex; c <= lci; c++) {
        d.width += columnViewManager.getSize(c);
        g.drawLine(d.width, d.height, d.width, lastY);
      }

      d.width = cr2.x;
      d.height = cr2.y;
      g.drawLine(d.width, d.height, lastX, d.height);
      for (int r = hit.rowIndex; r <= lri; r++) {
        d.height += rowViewManager.getSize(r);
        g.drawLine(d.width, d.height, lastX, d.height);
      }
    } else if (lastX >= 0 && lastY >= 0) {
        g.setColor(getBackground());
        g.fillRect(0, 0, vp.width, vp.height);
    } else {
      g.dispose();
      return;
    }

    pg.drawImage(canvas, clip.x, clip.y, null);
    g.dispose();
  }

  public void endEdit() throws Exception {
    endEdit(true);
  }

  public void endEdit(boolean post) throws Exception {
    ItemEditor editor = this.editor;
    this.editor = null;
    if (editor != null) {
      if (post) {
        try {
          if (canSet(editorLocation.rowIndex,
            editorLocation.columnIndex)) {

            model.set(editorLocation.rowIndex,
              editorLocation.columnIndex,
              editor.getValue());
          }

          editor.endEdit(post);
        } catch (Exception x) {
          this.editor = editor;
          throw x;
        }
      }

      editor.removeKeyListener(this);
      editor.removeKeyListener(keyCaster);
      Component editorComponent = editor.getComponent();
      if (editorComponent != null) {
        remove(editorComponent);
        editorComponent.removeFocusListener(this);
      }

      editorLocation = null;
      editClickPoint = null;
      requestFocus();
    }
  }

  public void safeEndEdit() {
    safeEndEdit(true);
  }

  public void safeEndEdit(boolean post) {
    try {
      endEdit(post);
    } catch (Exception ex) {}
  }

  public Point getEditClickPoint() {
    return editClickPoint;
  }
  
  public Dimension getPreferredSize() {
    int cCount = getColumnCount();
    int rCount = getRowCount();
    int width = columnViewManager.getSizeUpTo(cCount);
    int height = rowViewManager.getSizeUpTo(rCount);
    return new Dimension(width, height);
  }

  public void setBatchMode(boolean batchMode) {
    if (this.batchMode != batchMode) {
      this.batchMode = batchMode;
      if (!batchMode) {
        repaintCells();
        scroller.doLayout();
      }
    }
  }

  public boolean isBatchMode() {
    return batchMode;
  }

  public MatrixLocation hitTest(Point p) {
    return hitTest(p.x, p.y);
  }

  public MatrixLocation hitTest(int x, int y) {
    if (getRowCount() <= 0 || getColumnCount() <= 0) {
      return null;
    }

    Point adjust = getOriginPoint();

    int c = 0;
    int w = columnViewManager.getSize(c) - adjust.x;
    int cCount = getColumnCount();
    while (x >= w && ++c < cCount) {
      w += columnViewManager.getSize(c);
    }

    int r = 0;
    int h = rowViewManager.getSize(r) - adjust.y;
    int rCount = getRowCount();
    while (y >= h && ++r < rCount) {
      h += rowViewManager.getSize(r);
    }

    MatrixLocation hit = null;
    if (x >= -adjust.x && x < w && y >= -adjust.y && y < h) {
      hit = new MatrixLocation(r, c);
    }

    return hit;
  }

  public Rectangle getCellRect(int rowIndex, int columnIndex) {
    return getCellRect(new MatrixLocation(rowIndex, columnIndex));
  }

  public Rectangle getCellRect(MatrixLocation cell) {
    Point adjust = getOriginPoint();
    Rectangle rect = new Rectangle(-adjust.x, -adjust.y, 0, 0);
    rect.x += columnViewManager.getSizeUpTo(cell.columnIndex);
    rect.y += rowViewManager.getSizeUpTo(cell.rowIndex);
    rect.width = columnViewManager.getSize(cell.columnIndex);
    rect.height = rowViewManager.getSize(cell.rowIndex);
    return rect;
  }

  public int getRowCount() {
    return model == null ? 0 : model.getRowCount();
  }

  public int getColumnCount() {
    return model == null ? 0 : model.getColumnCount();
  }

  public MatrixLocation getSubfocus() {
    return subfocus;
  }

  public void setSubfocus(MatrixLocation newSubfocus) {
    setSubfocus(newSubfocus.rowIndex, newSubfocus.columnIndex);
  }

  public void setSubfocus(int rowIndex, int columnIndex) {
    setSubfocus(rowIndex, columnIndex, getFlags());
  }

  protected void setSubfocus(int rowIndex, int columnIndex,
    int flags) {

    if (batchMode || getRowCount() < 1 ||
      getColumnCount() < 1) {

      return;
    }

    if (rowIndex < 0 || columnIndex < 0 ||
      rowIndex > getRowCount() ||
      columnIndex > getColumnCount()) {

      throw new IllegalArgumentException(
        "rowIndex:" + rowIndex + ",columnIndex:" + columnIndex);
    }

    if (editor != null) {
      safeEndEdit();
    }

    requestFocus();

    if (!subfocusCaster.vetoableDispatch(
      new MatrixSubfocusEvent(
      this, SubfocusEvent.SUBFOCUS_CHANGING,
      new MatrixLocation(rowIndex, columnIndex)))) {

      return;
    }

    if (subfocus == null) {
      subfocus = new MatrixLocation(rowIndex, columnIndex);
      selectAnchor = new MatrixLocation(rowIndex, columnIndex);
      if ((flags & SelectFlags.ADD_ITEM) != 0) {
        selection.add(subfocus);
      }

      processSubfocusEvent(new MatrixSubfocusEvent(
        this, SubfocusEvent.SUBFOCUS_CHANGED,
        new MatrixLocation(subfocus)));

      return;
    }

    if (rowIndex != subfocus.rowIndex ||
      columnIndex != subfocus.columnIndex) {

      if (selectAnchor == null) {
        selectAnchor = new MatrixLocation(rowIndex, columnIndex);
      }

      MatrixLocation oldFocus = new MatrixLocation(subfocus);
      subfocus = new MatrixLocation(rowIndex, columnIndex);

      scrollView();

      if ((flags & SelectFlags.CLEAR) != 0) {
        selection.removeAll();
      }
      if ((flags & SelectFlags.ADD_ITEM) != 0) {
        selection.add(new MatrixLocation(subfocus));
      }

      if ((flags & SelectFlags.ADD_RANGE) != 0) {
        selection.removeRange(selectAnchor, oldFocus);
        selection.addRange(selectAnchor, subfocus);
      }

      if ((flags & SelectFlags.RESET_ANCHOR) != 0) {
        selectAnchor = new MatrixLocation(subfocus);
      }

      //repaintCell(oldFocus);
      //repaintCell(subfocus);

      processSubfocusEvent(new MatrixSubfocusEvent(
        this, SubfocusEvent.SUBFOCUS_CHANGED,
        new MatrixLocation(subfocus)));
    }
  }

  public MatrixLocation getOrigin() {
    return origin;
  }

  public synchronized void setOrigin(
    MatrixLocation origin) {

    Rectangle oldRect =
      this.origin != null ?
      getCellRect(this.origin) :
      getCellRect(0,0);

    Rectangle newRect =
      origin != null ?
      getCellRect(origin) :
      getCellRect(0,0);

    Dimension sz = getSize();
    int dx = oldRect.x - newRect.x;
    int dy = oldRect.y - newRect.y;

    this.origin = (origin != null &&
      !(origin.rowIndex == 0 &&
      origin.columnIndex == 0)) ?
      new MatrixLocation(origin) : null;

    if (batchMode) {
      return;
    }

    if (editor != null) {
      safeEndEdit();
    }

    if (dx == 0 && dy == 0) {
      return;
    }

    repaint();
  }

  public void setOrigin(int x, int y) {
    setOrigin(new MatrixLocation(x, y));
  }

  public Point getOriginPoint() {
    Point o = origin != null ?
      new Point(columnViewManager.getSizeUpTo(origin.columnIndex),
      rowViewManager.getSizeUpTo(origin.rowIndex)) :
      new Point(0, 0);

    return o;
  }

  public Rectangle getCellRangeRect(
    MatrixLocation start, MatrixLocation end) {

    MatrixLocation s = new MatrixLocation(
      Math.min(start.rowIndex, end.rowIndex),
      Math.min(start.columnIndex, end.columnIndex));

    MatrixLocation e = new MatrixLocation(
      Math.max(start.rowIndex, end.rowIndex),
      Math.max(start.columnIndex, end.columnIndex));

    if (s.rowIndex <= e.rowIndex &&
      s.columnIndex <= e.columnIndex) {

      Rectangle sr = getCellRect(s);
      if (sr != null) {
        Rectangle er = getCellRect(e);
        if (er != null) {
          return sr.union(er);
        } else {
          Dimension sz = getSize();
          Rectangle top = getCellRect(
            new MatrixLocation(0, e.columnIndex));

          int xmax = top != null ?
            top.x + top.width - 1 :
            sz.width - 1;

          Rectangle left = getCellRect(
            new MatrixLocation(e.rowIndex, 0));

          int ymax = left != null ?
            left.y + left.height - 1 :
            sz.height - 1;

          sr.add(xmax, ymax);
          return sr;
        }
      }
    }

    return null;
  }

  public void setDirectDraw(boolean directDraw) {
    this.directDraw = directDraw;
  }

  public boolean isDirectDraw() {
    return directDraw;
  }

  public void reset() {
    safeEndEdit(false);
    resize = null;
    setSubfocus(0, 0);
    requestFocus();
    scroller.setScrollPosition(0,0);
    repaintCells();
  }

  boolean startResize(
    boolean column, int index, int mouseX, int mouseY) {
    resize1 = column ?
      new MatrixLocation(0, index) :
      new MatrixLocation(index, 0);

    return startResize(column, resize1, mouseX, mouseY);
  }

  boolean startResize(boolean column,
    MatrixLocation location, int mouseX, int mouseY) {

    if (editor != null) {
      safeEndEdit();
    }

    Rectangle outerRect = new Rectangle(
      0, 0, getSize().width, getSize().height);

    Rectangle r = getCellRect(location);
    if (column) {
      dividerDelta = mouseX - (r.x + r.width);
      divider.setBounds(r.x + r.width, 0,
        dividerWidth, outerRect.height);
    } else {
      dividerDelta = mouseY - (r.y + r.height);
      divider.setBounds(0, r.y + r.height,
        outerRect.width, dividerHeight);
    }

    divider.setVisible(true);
    return true;
  }

  boolean whileResize(boolean column,
    int index, int mouseX, int mouseY) {

    if (column) {
      divider.setLocation(mouseX - dividerDelta, 0);
    } else {
      divider.setLocation(0, mouseY - dividerDelta);
    }

    return true;
  }

  boolean stopResize(boolean column,
    int index, int mouseX, int mouseY) {

    boolean retVal = stopResize(
      column, resize1, mouseX, mouseY);

    resize1 = null;
    invalidate();
    repaintCells();

    if (!batchMode) {
      scroller.doLayout();
    }

    return retVal;
  }

  boolean stopResize(boolean column,
    MatrixLocation location, int mouseX, int mouseY) {

    divider.setVisible(false);

    if (location != null) {
      Rectangle r = getCellRect(location);
      if (column) {
        if (mouseX - dividerDelta - r.x > 1) {
          columnViewManager.setSize(location.columnIndex,
            mouseX - dividerDelta - r.x);
        }
      } else {
        if (mouseY - dividerDelta - r.y > 1) {
          rowViewManager.setSize(location.rowIndex,
            mouseY - dividerDelta - r.y);
        }
      }
    }

    return true;
  }

  public ColumnViewManager getColumnViewManager() {
    return columnViewManager;
  }

  public RowViewManager getRowViewManager() {
    return rowViewManager;
  }

  public int getAlignment() {
    if (subfocus != null &&
      columnViewManager.getCount() > subfocus.columnIndex) {

      return columnViewManager.getAlignment(
        subfocus.columnIndex);
    } else {
      return super.getAlignment();
    }
  }

  public void addModelListener(
    MatrixModelListener listener) {

    modelCaster.add(listener);
  }

  public void removeModelListener(
    MatrixModelListener listener) {

    modelCaster.remove(listener);
  }

  public void addSelectionListener(
    MatrixSelectionListener listener) {

    selectionCaster.add(listener);
  }

  public void removeSelectionListener(
    MatrixSelectionListener listener) {

    selectionCaster.remove(listener);
  }

  public void addSubfocusListener(
    MatrixSubfocusListener listener) {

    subfocusCaster.add(listener);
  }

  public void removeSubfocusListener(
    MatrixSubfocusListener listener) {

    subfocusCaster.remove(listener);
  }

  public void addKeyListener(KeyListener listener) {
    keyCaster.add(listener);
  }

  public void removeKeyListener(KeyListener listener) {
    keyCaster.remove(listener);
  }

  public void modelContentChanged(MatrixModelEvent e) {
    MatrixLocation ml = e.getLocation();
    
    switch (e.getChange()) {
      case MatrixModelEvent.CONTENT_CHANGED:
        repaintCells();
        break;
      case MatrixModelEvent.ITEM_CHANGED:
        if (editor != null && editorLocation != null &&
          editorLocation.equals(ml)) {

          safeEndEdit(false);
        }

        repaintCell(ml);
        break;
      case MatrixModelEvent.ROW_CHANGED:
        if (editor != null && editorLocation != null &&
          editorLocation.rowIndex == ml.rowIndex) {
          safeEndEdit(false);
        }

        repaintCells(new MatrixLocation(ml.rowIndex, 0),
          new MatrixLocation(getRowCount() - 1,
          getColumnCount() - 1));

        break;
      case MatrixModelEvent.COLUMN_CHANGED:
        if (editor != null && editorLocation != null &&
          editorLocation.columnIndex == ml.columnIndex) {

          safeEndEdit(false);
        }

        repaintCells(new MatrixLocation(0, ml.columnIndex),
          new MatrixLocation(getRowCount() - 1,
          getColumnCount() - 1));

        break;
    }
  }

  public void modelStructureChanged(MatrixModelEvent e) {
    MatrixLocation ml = e.getLocation();

    switch (e.getChange()) {
      case MatrixModelEvent.STRUCTURE_CHANGED:
        if (editor != null) {
          safeEndEdit(false);
        }

        invalidate();
        repaintCells();
        if (!batchMode) {
          scroller.doLayout();
        }

        break;
      case MatrixModelEvent.ROW_ADDED:
      case MatrixModelEvent.ROW_REMOVED:
        if (editor != null) {
          safeEndEdit();
        }

        invalidate();
        if (ml != null && subfocus != null) {
          setSubfocus(ml.rowIndex, subfocus.columnIndex);
        }

        repaintCells();
        if (!batchMode) {
          scroller.doLayout();
        }

        break;
      case MatrixModelEvent.COLUMN_ADDED:
        columnViewManager.addColumnView(ml.columnIndex);
        invalidate();
        repaintCells();
        if (!batchMode) {
          scroller.doLayout();
        }

        break;
      case MatrixModelEvent.COLUMN_REMOVED:
        int cols = getColumnCount();
        if (subfocus != null &&
          cols <= subfocus.columnIndex && cols > 0) {
          setSubfocus(subfocus.rowIndex, cols - 1);
        }

        columnViewManager.dropColumnView(ml.columnIndex);
        invalidate();
        repaintCells();
        if (!batchMode) {
          scroller.doLayout();
        }

        break;
    }
  }

  public void selectionItemChanged(MatrixSelectionEvent e) {
    repaintCell(e.getLocation());
    oldSelected = selection.getAll();
  }

  public void selectionRangeChanged(MatrixSelectionEvent e) {
    switch (e.getChange()) {
      case SelectionEvent.RANGE_ADDED:
      case SelectionEvent.RANGE_REMOVED:
        selectionChanged(e);
        break;
    }
  }

  public void selectionChanged(MatrixSelectionEvent e) {
    switch (e.getChange()) {
      case SelectionEvent.SELECTION_CLEARED:
        repaintSelection();
        break;
      case SelectionEvent.SELECTION_CHANGED:
      default:
        MatrixLocation[] selected = selection.getAll();
        MatrixSelection old =
          new BasicMatrixSelection(oldSelected);

        setDirectDraw(false);
        for (int i = 0 ; i < selected.length ; i++) {
          if (!old.contains(selected[i])) {
            repaintCell(selected[i]);
          } else {
            old.remove(selected[i]);
          }
        }

        oldSelected = old.getAll();
        repaintSelection();
        setDirectDraw(true);
        break;
    }

    oldSelected = selection.getAll();
  }

  public void keyPressed(KeyEvent e) {
    processKeyPressed(e);
  }

  public void keyReleased(KeyEvent e) {}

  public void keyTyped(KeyEvent e) {};

  protected void processKeyPressed(KeyEvent e) {
    super.processKeyPressed(e);
    if (e.isConsumed()) {
      return;
    }

    boolean alt = e.isAltDown();
    boolean control = e.isControlDown();
    boolean shift = e.isShiftDown();
    int flags = getFlags(shift, control);

    if (alt) {
      return;
    }

    switch (e.getKeyCode()) {
      case KeyEvent.VK_ENTER:
        if (control && !shift && editor == null &&
          canSet(subfocus.rowIndex, subfocus.columnIndex)) {

          startEdit(subfocus);
        } else if (editor != null) {
          safeEndEdit();
        }

        e.consume();
        break;
      case KeyEvent.VK_F2:
        if ((editor == null) &&
          canSet(subfocus.rowIndex, subfocus.columnIndex)) {

          startEdit(subfocus);
        }

        e.consume();
        break;
      case KeyEvent.VK_ESCAPE:
        safeEndEdit(false);
        e.consume();
        break;
      case KeyEvent.VK_DOWN:
        if (subfocus.rowIndex < getRowCount() - 1) {
          moveFocus(1, 0, flags);
        } else {
          safeEndEdit();
        }

        e.consume();
        break;
      case KeyEvent.VK_UP:
        if (subfocus.rowIndex > 0) {
          moveFocus(-1, 0, flags);
        } else {
          safeEndEdit();
        }

        e.consume();
        break;
      case KeyEvent.VK_PAGE_UP:
        if (!control && subfocus.rowIndex > 0) {
          pageJump(false, control, flags);
        } else if (control &&
          subfocus.columnIndex > 0) {

          flags = getFlags(shift, false);
          pageJump(false, control, flags);
        }

        e.consume();
        break;
      case KeyEvent.VK_PAGE_DOWN:
        if (!control &&
          subfocus.rowIndex < getRowCount() - 1) {

          pageJump(true, control, flags);
        } else if (control &&
          subfocus.columnIndex < getColumnCount() - 1) {

          flags = getFlags(shift, false);
          pageJump(true, control, flags);
        }

        e.consume();
        break;
      case KeyEvent.VK_LEFT:
        if (editor == null &&
          subfocus.columnIndex > 0) {

          moveFocus(0, -1, flags);
          e.consume();
        }

        break;
      case KeyEvent.VK_RIGHT:
        if (editor == null &&
          subfocus.columnIndex < getColumnCount() - 1) {

          moveFocus(0, 1, flags);
          e.consume();
        }

        break;
      case KeyEvent.VK_TAB:
        flags = getFlags();
        if (!shift) {
          if (subfocus.columnIndex < getColumnCount() - 1) {
            moveFocus(0, 1, flags);
          } else if (subfocus.rowIndex < getRowCount() - 1) {
            moveFocus(1, -subfocus.columnIndex, flags);
          } else {
            moveFocus(-subfocus.rowIndex,
              -subfocus.columnIndex, flags);
          }
        } else {
          if (subfocus.columnIndex > 0) {
            moveFocus(0, -1, flags);
          } else if (subfocus.rowIndex > 0) {
            moveFocus(-1, getColumnCount() - 1, flags);
          } else {
            moveFocus(getRowCount() - 1,
              getColumnCount() - 1, flags);
          }
        }

        e.consume();
        break;
      case KeyEvent.VK_SPACE:
        if (control && !shift) {
          if (selection.contains(subfocus)) {
            selection.remove(subfocus);
          } else {
            selection.add(subfocus);
          }

          e.consume();
        }

        break;
      case KeyEvent.VK_HOME:
        if (control && shift) {
          if (subfocus.rowIndex != 0 ||
            subfocus.columnIndex != 0) {

            setSubfocus(0, 0, SelectFlags.ADD_RANGE);
          }
        } else if (control) {
          if (subfocus.rowIndex != 0 ||
            subfocus.columnIndex != 0) {

            setSubfocus(0, 0);
          }
        } else if (subfocus.columnIndex != 0) {
          setSubfocus(subfocus.rowIndex, 0, flags);
        }

        e.consume();
        break;
      case KeyEvent.VK_END:
        if (control && shift) {
          if (subfocus.rowIndex != getRowCount() - 1 ||
            subfocus.columnIndex != getColumnCount()-1) {

            setSubfocus(getRowCount() - 1,
              getColumnCount() - 1, SelectFlags.ADD_RANGE);
          }
        } else if (control) {
          if (subfocus.rowIndex != getRowCount() - 1 ||
            subfocus.columnIndex != getColumnCount() - 1) {
            setSubfocus(getRowCount() - 1,
              getColumnCount() - 1);
          }
        } else if (subfocus.columnIndex != getColumnCount() - 1) {
          setSubfocus(subfocus.rowIndex,
            getColumnCount() - 1, flags);
        }

        e.consume();
        break;
      case KeyEvent.VK_INSERT:
        if (editor != null || isReadOnly()) {
          break;
        }

        if (shift && control) {
          model.addColumn(subfocus.columnIndex);
        } else if (control) {
          model.addRow(subfocus.rowIndex);
        }

        e.consume();
        break;
      case KeyEvent.VK_DELETE:
        if (editor != null || isReadOnly()) {
          break;
        }

        if (shift && control) {
          if (getColumnCount() > 0) {
            model.removeColumn(subfocus.columnIndex);
            if (subfocus.columnIndex == getColumnCount()) {
              moveFocus(0, -1, getFlags());
            }
          }
        } else if (control) {
          if (getRowCount() > 0) {
            model.removeRow(subfocus.rowIndex);
            if (subfocus.rowIndex == getRowCount()) {
              moveFocus(-1, 0, getFlags());
            }
          }
        }

        e.consume();
        break;
      case KeyEvent.VK_D:
        if (control && editor == null &&
          subfocus.rowIndex > 0 &&
          canSet(subfocus.rowIndex, subfocus.columnIndex)) {

          model.set(subfocus.rowIndex, subfocus.columnIndex,
            model.get(subfocus.rowIndex - 1, subfocus.columnIndex));

          e.consume();
        }

        break;
    }
  }

  protected void processMousePressed(MouseEvent e) {
    super.processMousePressed(e);
    if (rangeSelecting) {
      return;
    }

    int x = e.getX();
    int y = e.getY();
    boolean shift = e.isShiftDown();
    boolean control = e.isControlDown();
    boolean right = e.isMetaDown();
    int flags = getFlags(shift, control);

    MatrixLocation current = hitTest(x, y);
    mouseDown = current != null ?
      new MatrixLocation(current) : null;

    if (editor != null) {
      if (editorLocation.equals(current)) {
        return;
      } else {
        safeEndEdit();
      }
    }

    resize = null;
    if (current != null) {
      if (current.equals(subfocus)) {
        if (!right && !control && !shift &&
          canSet(current.rowIndex, current.columnIndex)) {
          
          editClickPoint = new Point(x, y);
          startEdit(subfocus);
          return;
        }
      }

      if (control && !shift) {
        if (selection.contains(current)) {
          selection.remove(current);
        } else {
          selection.add(new MatrixLocation(current));
        }
      }

      rangeSelecting = true;
      setSubfocus(current.rowIndex,
        current.columnIndex, flags);
    }
  }

  protected void processMouseReleased(MouseEvent e) {
    super.processMouseReleased(e);
    rangeSelecting = false;
    mouseDown = null;
  }

  protected void processMouseDragged(MouseEvent e) {
    super.processMouseDragged(e);
    if (!e.isMetaDown() && rangeSelecting) {
      int x = e.getX();
      int y = e.getY();
      int flags = e.isControlDown() ?
        SelectFlags.RESET_ANCHOR :
        SelectFlags.ADD_RANGE;

      MatrixLocation current = hitTest(x, y);
      if (current != null) {
        setSubfocus(current.rowIndex,
          current.columnIndex, flags);
      }
    }
  }

  public void focusGained(FocusEvent e) {
    processFocusGained(e);
  }

  public void focusLost(FocusEvent e) {
    processFocusLost(e);
  }

  protected void processFocusGained(FocusEvent e) {
    super.processFocusGained(e);
    if (editor != null &&
      editor.getComponent() != null) {

      editor.getComponent().requestFocus();
    }

    if (!hasFocus) {
      hasFocus = true;
      repaintCell(subfocus);
    }
  }

  protected void processFocusLost(FocusEvent e) {
    super.processFocusLost(e);
    if (hasFocus) {
      hasFocus = false;
      repaintCell(subfocus);
    }
  }

  protected void processSubfocusEvent(MatrixSubfocusEvent e) {
    subfocusCaster.dispatch(e);
  }

  protected void moveFocus(
    int deltaRows, int deltaColumns, int flags) {

    int r = Math.max(0, Math.min(
      subfocus.rowIndex + deltaRows, getRowCount() - 1));

    int c = Math.max(0, Math.min(
      subfocus.columnIndex + deltaColumns, getColumnCount() - 1));

    setSubfocus(r, c, flags);
  }

  protected void startEdit(MatrixLocation newEditorLocation) {
    if (model == null || batchMode ||
      !canSet(newEditorLocation.rowIndex,
      newEditorLocation.columnIndex) ) {

      return;
    }

    editorLocation = new MatrixLocation(newEditorLocation);
    selection.removeAll();
    selection.add(new MatrixLocation(editorLocation));

    Object data = model.get(
      editorLocation.rowIndex, editorLocation.columnIndex);

    int state = getState(
      editorLocation.rowIndex, editorLocation.columnIndex);

    editor = getEditor(editorLocation.columnIndex);
    if (editor != null) {
      Component editorComponent = editor.getComponent();
      if (editorComponent != null) {
        editorComponent.setVisible(false);
        add(editorComponent);
      }

      Rectangle r = getCellRect(editorLocation);
      r.x--;
      r.y--;
      r.width++;
      r.height++;

      ItemEditSite site = this;
      if (columnViewManager.getCount() >
        editorLocation.columnIndex) {

        ColumnView cv = columnViewManager.getColumnView(
          editorLocation.columnIndex);

        cv.setEditClickPoint(editClickPoint);
        site = cv;
      }

      editor.addKeyListener(this);
      editor.addKeyListener(keyCaster);
      editor.startEdit(data, r, site);
      if (editorComponent != null) {
        editorComponent.addFocusListener(this);
      }

      editClickPoint = null;
    }
  }

  protected int getLastColumnIndex() {
    Dimension vp = getSize();
    MatrixLocation topRight =
        hitTest(vp.width - 1, 0);

    int lastColumnIndex =
      topRight != null ? topRight.columnIndex :
      getColumnCount() - 1;

    return lastColumnIndex;
  }

  protected int getLastRowIndex() {
    Dimension vp = getSize();
    MatrixLocation leftBottom =
      hitTest(0, vp.height - 1);

    int lastRowIndex =
      leftBottom != null ? leftBottom.rowIndex :
      getRowCount() - 1;

    return lastRowIndex;
  }

  protected int getOriginRow() {
    int r = origin != null ?
      origin.rowIndex : 0;

    return r;
  }

  protected int getOriginColumn() {
    int c = origin != null ?
      origin.columnIndex : 0;

    return c;
  }

  protected int getState(int rowIndex, int columnIndex) {
    int state = 0;
    if (selection.contains(rowIndex, columnIndex)) {
      state |= ItemPainter.SELECTED;
      if (selectAnchor.rowIndex == rowIndex &&
        selectAnchor.columnIndex == columnIndex) {
        state |= ItemPainter.ANCHOR;
      }
    }

    if (!isEnabled()) {
      state |= ItemPainter.DISABLED;
    } else {
      if (hasFocus && subfocus.rowIndex == rowIndex &&
        subfocus.columnIndex == columnIndex) {
        state |= ItemPainter.FOCUSED;
      }
    }
    return state;
  }

  private boolean canSet(int rowIndex, int columnIndex) {
    return isReadOnly() ? false :
      model.canSet(rowIndex, columnIndex);
  }

  private int getFlags() {
    return getFlags(false, false);
  }

  private int getFlags(boolean shift, boolean control) {
    int flags;
    if (shift && control) {
      flags = SelectFlags.ADD_RANGE;
    } else if (shift) {
      flags = SelectFlags.CLEAR |
        SelectFlags.ADD_RANGE;
    } else if (control) {
      flags = SelectFlags.RESET_ANCHOR |
        SelectFlags.ADD_ITEM;
    } else {
      flags = SelectFlags.CLEAR | SelectFlags.ADD_ITEM |
      SelectFlags.RESET_ANCHOR;
    }
    
    return flags;
  }

  private void scrollView() {
    MatrixLocation newOrigin = origin != null ?
      new MatrixLocation(origin) :
      new MatrixLocation();

    boolean scroll = false;

    Dimension sz = getSize();
    Rectangle fRect = getCellRect(subfocus);
    if (fRect != null) {
      if (fRect.x < 0) {
        scroll = true;
        newOrigin.columnIndex = subfocus.columnIndex;
      } else if (sz.width < fRect.x + fRect.width) {
        scroll = true;
        MatrixLocation hit = hitTest(
          fRect.x + fRect.width - sz.width,
          fRect.y);
        if (hit != null) {
          newOrigin.columnIndex = hit.columnIndex + 1;
        } else {
          return;
        }
      }

      if (fRect.y < 0) {
        scroll = true;
        newOrigin.rowIndex = subfocus.rowIndex;
      } else if (sz.height < fRect.y + fRect.height) {
        scroll = true;
        MatrixLocation hit = hitTest(
        fRect.x,
        fRect.y + fRect.height - sz.height);

        if (hit != null) {
          newOrigin.rowIndex = hit.rowIndex + 1;
        } else {
          return;
        }
      }

      if (scroll) {
        scroller.setScrollPosition(
          newOrigin.columnIndex, newOrigin.rowIndex);
      }
    }
  }

  private void pageJump(boolean pageDown,
    boolean control,int flags) {
    safeEndEdit();
    int dx = !control ?
      getLastRowIndex() - getOriginRow() : 0;

    int dy = !control ?
      0 : getLastColumnIndex() - getOriginColumn();

    dx = pageDown ? dx : -dx;
    dy = pageDown ? dy : -dy;
    moveFocus(dx, dy, flags);
  }
}
