package higa.jhcl.view;

import java.awt.SystemColor;
import java.awt.Scrollbar;
import java.awt.Rectangle;
import java.awt.Graphics;
import java.awt.FontMetrics;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Point;

import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.text.Format;

import java.io.Serializable;

import higa.jhcl.model.ItemPainter;
import higa.jhcl.model.ItemEditor;
import higa.jhcl.model.MatrixLocation;
import higa.jhcl.model.MatrixModel;
import higa.jhcl.model.MatrixModelListener;
import higa.jhcl.model.MatrixSubfocusEvent;
import higa.jhcl.model.MatrixSubfocusListener;
import higa.jhcl.model.MatrixSelection;
import higa.jhcl.model.MatrixSelectionEvent;
import higa.jhcl.model.MatrixSelectionListener;

import higa.jhcl.control.HPanel;

import higa.util.Orientation;
import higa.util.Alignment;
import higa.util.Formatter;

public class GridView extends HPanel
  implements Scroller, HeaderListener,
  Serializable {

  protected GridCore core = new GridCore(this);
  protected HComponent panel = new HComponent();
  protected ButtonView button =
    new ButtonView();

  protected Scrollbar vScroll =
    new Scrollbar(Scrollbar.VERTICAL);

  protected int vScrollWidth = 16;
  protected boolean showVScroll = true;

  protected Scrollbar hScroll =
    new Scrollbar(Scrollbar.HORIZONTAL);

  protected int hScrollHeight = 16;
  protected boolean showHScroll = true;

  protected HeaderView columnHeader =
    new HeaderView(Orientation.HORIZONTAL);

  protected HeaderView rowHeader =
    new HeaderView(Orientation.VERTICAL);

  public GridView() {
    init();
  }

  public void addNotify() {
    super.addNotify();
    repaintCells();
  }

  private void init() {
    setBevelOuter(LOWERED);
    setBevelInner(LOWERED);
    panel.setBackground(SystemColor.control);

    vScroll.addAdjustmentListener(
      new AdjustmentListener() {
        public void adjustmentValueChanged(AdjustmentEvent e) {
          scrolled(true, e.getValue());
        }
      });

    hScroll.addAdjustmentListener(
      new AdjustmentListener() {
        public void adjustmentValueChanged(AdjustmentEvent e) {
          scrolled(false, e.getValue());
        }
      });

    columnHeader.addHeaderListener(this);
    ColumnViewManager cvm = core.getColumnViewManager();
    columnHeader.setModel(cvm);
    columnHeader.setItemSizes(cvm);

    rowHeader.addHeaderListener(this);
    RowViewManager rvm = core.getRowViewManager();
    rowHeader.setModel(rvm);
    rowHeader.setItemSizes(rvm);

    button.setFocusAware(false);
    button.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          core.reset();
        }
      });

    add(core);
    add(columnHeader);
    add(rowHeader);
    add(button);
    add(vScroll);
    add(hScroll);
    add(panel);
  }

  public void scrolled(boolean vert, int value) {
    MatrixLocation oldOrigin = core.getOrigin();
    if (oldOrigin == null) {
      oldOrigin = new MatrixLocation();
    }

    MatrixLocation newOrigin = new MatrixLocation(oldOrigin);
    if (vert) {
      newOrigin.rowIndex = value;
      rowHeader.setOrigin(value);
    } else {
      newOrigin.columnIndex = value;
      columnHeader.setOrigin(value);
    }

    core.setOrigin(newOrigin);
  }

  public void doLayout() {
    Dimension sz = getSize();
    Dimension ps = core.getPreferredSize();
    Insets bInsets = getInsets();
    if (sz == null || ps == null || bInsets == null) {
      return;
    }

    int chh = (columnHeader == null ||
      !columnHeader.isVisible()) ? 0 :
      columnHeader.getThickness();

    int rhw = (rowHeader == null ||
      !rowHeader.isVisible()) ? 0 :
      rowHeader.getThickness();

    boolean vsv =
      ((bInsets.top + chh + ps.height + bInsets.bottom) >
      sz.height) && showVScroll;

    boolean hsv =
      ((bInsets.left + rhw + ps.width + bInsets.right) >
      sz.width) && showHScroll;

    if (!vsv && hsv && showVScroll) {
      vsv = (bInsets.top + chh + ps.height +
        hScrollHeight + bInsets.bottom) > sz.height;
    }

    if (!hsv && vsv && showHScroll) {
      hsv = (bInsets.left + rhw + ps.width +
      vScrollWidth + bInsets.right) > sz.width;
    }

    int vsw = vsv ? vScrollWidth : 0;
    int hsh = hsv ? hScrollHeight : 0;
    core.setBounds(
      bInsets.left + rhw, bInsets.top + chh,
      sz.width - bInsets.left - rhw - vsw - bInsets.right,
      sz.height - bInsets.top - chh - hsh - bInsets.bottom);

    vScroll.setVisible(vsv);
    if (vsv) {
      vScroll.setBounds(
        sz.width - vsw - bInsets.right, bInsets.top,
        vsw, sz.height - hsh - bInsets.top - bInsets.bottom);
    }

    hScroll.setVisible(hsv);
    if (hsv) {
      hScroll.setBounds(
        bInsets.left, sz.height - hsh - bInsets.bottom,
        sz.width - vsw - bInsets.left - bInsets.right, hsh);
    }

    if (chh > 0) {
      columnHeader.setBounds(
        rhw + bInsets.left, bInsets.top,
        sz.width - rhw - vsw - bInsets.left - bInsets.right, chh);
    }

    if (rhw > 0) {
      rowHeader.setBounds(
        bInsets.left, chh + bInsets.top,
        rhw, sz.height - chh - hsh - bInsets.top - bInsets.bottom);
    }

    if (chh > 0 && rhw > 0) {
      button.setBounds(bInsets.left, bInsets.right, rhw, chh);
      button.setVisible(true);
    } else {
      button.setVisible(false);
    }

    panel.setVisible(vsv && hsv);
    if (vsv && hsv) {
      panel.setBounds(
        sz.width - bInsets.right - vsw,
        sz.height - bInsets.bottom - hsh,
        vsw, hsh);
    }

    adjustScrollbars();
  }

  protected void adjustScrollbars() {
    Dimension ps = core.getPreferredSize();
    Dimension cs = core.getSize();

    if (ps == null || cs == null) {
      return;
    }

    int rowCount = core.getRowCount();
    int columnCount = core.getColumnCount();

    int visibleRowCount = 1;
    int visibleColumnCount = 1;
    MatrixLocation origin = core.getOrigin();
    if (origin == null) {
      origin = new MatrixLocation();
    }

    int h = 0;
    RowViewManager rvm = core.getRowViewManager();
    for (int r = 0; r < rowCount; r++) {
      h += rvm.getSize(r);
      if (h > cs.height) {
        visibleRowCount = r - 1;
        break;
      }
    }

    int w = 0;
    ColumnViewManager cvm = core.getColumnViewManager();
    for (int c = 0; c < columnCount; c++) {
      w += cvm.getSize(c);
      if (w > cs.width) {
        visibleColumnCount = c - 1;
        break;
      }
    }

    if (ps.height > cs.height) {
      if (!vScroll.isEnabled()) {
        vScroll.setEnabled(true);
      }

      int v = vScroll.getValue();
      vScroll.setMinimum(0);
      vScroll.setMaximum(rowCount);
      vScroll.setVisibleAmount(visibleRowCount);
      vScroll.setBlockIncrement(visibleRowCount);
      if (rowHeader.isVisible()) {
        rowHeader.setOrigin(origin.rowIndex);
      }
    }

    if (ps.width > cs.width) {
      if (!hScroll.isEnabled()) {
        hScroll.setEnabled(true);
      }

      int v = hScroll.getValue();
      hScroll.setMinimum(0);
      hScroll.setMaximum(columnCount);
      hScroll.setVisibleAmount(visibleColumnCount);
      hScroll.setBlockIncrement(visibleColumnCount);
      if (columnHeader.isVisible()) {
        columnHeader.setOrigin(origin.columnIndex);
      }
    }
  }

  public void update(Graphics g) {
    paint(g);
  }

  public void paint(Graphics g) {
    super.paint(g);
    Dimension sz = getSize();
    g.clipRect(0, 0, sz.width, sz.height);
  }

  public Dimension getViewportSize() {
    return core.getSize();
  }

  public void setScrollPosition(int x, int y) {
    hScroll.setValue(x);
    scrolled(false, x);
    vScroll.setValue(y);
    scrolled(true, y);
    adjustScrollbars();
  }
  
  public Point getScrollPosition() {
    ColumnViewManager cvm = core.getColumnViewManager();
    int hIndex = hScroll.getValue();
    RowViewManager rvm = core.getRowViewManager();
    int vIndex = vScroll.getValue();
    Point sp = new Point(cvm.getSizeUpTo(hIndex),
      rvm.getSizeUpTo(vIndex));
      
    return sp;
  }

  public ColumnViewManager getColumnViewManager() {
    return core.getColumnViewManager();
  }

  public RowViewManager getRowViewManager() {
    return core.getRowViewManager();
  }

  public void setColumnView(int index, ColumnView cv) {
    ColumnViewManager cvm = core.getColumnViewManager();
    cvm.setColumnView(index, cv);
    invalidate();
  }

  public ColumnView getColumnView(int index) {
    ColumnViewManager cvm = core.getColumnViewManager();
    return cvm.getColumnView(index);
  }

  public ColumnView getDefaultColumnView(int index) {
    ColumnViewManager cvm = core.getColumnViewManager();
    return cvm.getDefaultColumnView(index);
  }

  public void setShowVScroll(boolean visible) {
    showVScroll = visible;
    invalidate();
    if (isShowing()) {
      doLayout();
    }
  }

  public boolean isShowVScroll() {
    return showVScroll;
  }

  public void setShowHScroll(boolean visible) {
    showHScroll = visible;
    invalidate();
    if (isShowing()) {
      doLayout();
    }
  }

  public boolean isShowHScroll() {
    return showHScroll;
  }

  public Dimension getPreferredSize() {
    Dimension ps = core.getPreferredSize();

    ps.width += (showVScroll ? vScrollWidth : 0);
    ps.height += (showHScroll ? hScrollHeight : 0);
    return ps;
  }

  public void setModel(MatrixModel model) {
    invalidate();
    core.setModel(model);
    adjustColumnWidth();
  }

  private void adjustColumnWidth() {
    int columnCount = getColumnCount();
    int rowCount = Math.min(getRowCount(), 10);
    ColumnViewManager cvm = getColumnViewManager();
    cvm.setCount(columnCount);
    Graphics g = getGraphics();
    if (g == null) {
      return;
    }

    FontMetrics fm = g.getFontMetrics();
    if (fm == null) {
      return;
    }

    MatrixModel model = getModel();
    if (model == null) {
      return;
    }

    for (int c = 0; c < columnCount; c++) {
      ColumnView cv = getColumnView(c);
      String caption = cv.getCaption();
      ItemPainter ip = cv.getPainter();
      ip = ip != null ? ip : getPainter();
      int width = fm.stringWidth(caption);
      for (int r = 0; r < rowCount; r++) {
        Dimension d =ip.getPreferredSize(
          model.get(r, c), g, core.getState(r, c), cv);

        width = Math.max(width, d.width);
      }

      cv.setWidth(width + 10);
      if (ip instanceof ChoiceItemPainterEditor) {
        cv.setAlignment(Alignment.LEFT);
      }
      
      setColumnView(c, cv);
    }
  }

  public MatrixModel getModel() {
    return core.getModel();
  }

  public void setPainter(ItemPainter painter) {
    core.setPainter(painter);
  }

  public ItemPainter getPainter() {
    return core.getPainter();
  }

  public void setEditor(ItemEditor editor) {
    core.setEditor(editor);
  }

  public ItemEditor getEditor() {
    return core.getEditor();
  }

  public void setReadOnly(boolean ro) {
    core.setReadOnly(ro);
  }

  public boolean isReadOnly() {
    return core.isReadOnly();
  }

  public void setSelection(MatrixSelection ms) {
    core.setSelection(ms);
  }

  public MatrixSelection getSelection() {
    return core.getSelection();
  }

  public void setSubfocus(MatrixLocation newSubfocus) {
    core.setSubfocus(newSubfocus);
  }

  public MatrixLocation getSubfocus() {
    return core.getSubfocus();
  }

  public void setSubfocus(int rowIndex, int columnIndex) {
    core.setSubfocus(rowIndex, columnIndex);
  }

  public void requestFocus() {
    core.requestFocus();
  }

  public void repaintCell(MatrixLocation cell) {
    core.repaintCell(cell);
  }

  public void repaintCells() {
    core.repaintCells();
  }

  public void repaintCells(
    MatrixLocation start, MatrixLocation end) {
    core.repaintCells(start, end);
  }

  public void setBatchMode(boolean batchMode) {
    core.setBatchMode(batchMode);
    columnHeader.setBatchMode(batchMode);
    rowHeader.setBatchMode(batchMode);
  }

  public boolean isBatchMode() {
    return core.isBatchMode();
  }

  public MatrixLocation hitTest(int x, int y) {
    return core.hitTest(x, y);
  }

  public Rectangle getCellRect(MatrixLocation cell) {
    return core.getCellRect(cell);
  }

  public Rectangle getCellRangeRect(
    MatrixLocation start, MatrixLocation end) {

    return core.getCellRangeRect(start, end);
  }

  public int getColumnCount() {
    return core.getColumnCount();
  }

  public int getRowCount() {
    return core.getRowCount();
  }

  public void setRowHeight(int rowIndex, int height) {
    RowViewManager rvm = core.getRowViewManager();
    rvm.setSize(rowIndex, height);
    invalidate();
  }

  public int getRowHeight(int rowIndex) {
    RowViewManager rvm = core.getRowViewManager();
    return rvm.getSize(rowIndex);
  }

  public void setColumnWidth(int columnIndex, int width) {
    ColumnViewManager cvm = core.getColumnViewManager();
    cvm.setSize(columnIndex, width);
    invalidate();
  }

  public int getColumnWidth(int columnIndex) {
    ColumnViewManager cvm = core.getColumnViewManager();
    return cvm.getSize(columnIndex);
  }

  public void setDefaultRowHeight(int height) {
    RowViewManager rvm = core.getRowViewManager();
    rvm.setDefaultRowHeight(height);
  }

  public int getDefaultRowHeight() {
    RowViewManager rvm = core.getRowViewManager();
    return rvm.getDefaultRowHeight();
  }

  public void setRowHeaderWidth(int width) {
    if (rowHeader == null) {
      return;
    }

    invalidate();
    rowHeader.setThickness(width);
    core.repaintCells();
  }

  public int getRowHeaderWidth() {
    return rowHeader.getThickness();
  }

  public void setRowHeaderVisible(boolean visible) {
    if (rowHeader == null) {
      return;
    }

    if (visible != rowHeader.isVisible()) {
      invalidate();
      rowHeader.setVisible(visible);

      doLayout();
      core.repaintCells();
    }
  }

  public boolean isRowHeaderVisible() {
    return rowHeader.isVisible();
  }

  public void setColumnHeaderHeight(int height) {
    if (columnHeader == null) {
      return;
    }

    invalidate();
    columnHeader.setThickness(height);
    core.repaintCells();
  }

  public int getColumnHeaderHeight() {
    return columnHeader.getThickness();
  }

  public void addHeaderListener(HeaderListener l) {
    rowHeader.addHeaderListener(l);
    columnHeader.addHeaderListener(l);
  }

  public void removeHeaderListener(HeaderListener l) {
    rowHeader.removeHeaderListener(l);
    columnHeader.removeHeaderListener(l);
  }

  public void setColumnHeaderVisible(boolean visible) {
    if (columnHeader == null) {
      return;
    }

    if (visible != columnHeader.isVisible()) {
      invalidate();
      columnHeader.setVisible(visible);
      doLayout();
      core.repaintCells();
    }
  }

  public boolean isColumnHeaderVisible() {
    return columnHeader.isVisible();
  }

  protected void columnHeaderClicked(int index) {
    core.setSubfocus(core.getSubfocus().rowIndex, index);
  }

  protected void rowHeaderClicked(int index) {
    core.setSubfocus(index, core.getSubfocus().columnIndex);
  }

  public void headerItemClicked(HeaderEvent e) {
    Object source = e.getSource();
    int index = e.getIndex();
    if (source == columnHeader) {
      columnHeaderClicked(index);
    } else {
      rowHeaderClicked(index);
    }
  }

  public void headerItemResizing(HeaderEvent e) {
    Object source = e.getSource();
    int id = e.getID();
    int index = e.getIndex();
    int x = e.getX();
    int y = e.getY();
    switch (id) {
      case HeaderEvent.START_RESIZE:
        core.startResize(source == columnHeader, index, x, y);
        break;
      case HeaderEvent.WHILE_RESIZE:
        core.whileResize(source == columnHeader, index, x, y);
        break;
      case HeaderEvent.STOP_RESIZE:
        core.stopResize(source == columnHeader, index, x, y);
        break;
    }
  }

  public void addModelListener(
    MatrixModelListener listener) {

    core.addModelListener(listener);
  }

  public void removeModelListener(
    MatrixModelListener listener) {

    core.removeModelListener(listener);
  }

  public void addSelectionListener(
    MatrixSelectionListener listener) {

    core.addSelectionListener(listener);
  }
  
  public void removeSelectionListener(
    MatrixSelectionListener listener) {

    core.removeSelectionListener(listener);
  }

  public void addSubfocusListener(
    MatrixSubfocusListener listener) {

    core.addSubfocusListener(listener);
  }

  public void removeSubfocusListener(
    MatrixSubfocusListener listener) {

    core.removeSubfocusListener(listener);
  }
}
