package higa.jhcl.view;

import java.awt.Cursor;
import java.awt.SystemColor;
import java.awt.Graphics;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Image;
import java.awt.Container;

import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import higa.util.Orientation;
import higa.util.Alignment;
import higa.util.EventMulticaster;

import higa.jhcl.model.VectorModel;
import higa.jhcl.model.VectorModelEvent;
import higa.jhcl.model.VectorModelListener;
import higa.jhcl.model.BasicVectorContainer;
import higa.jhcl.model.ItemPainter;

public class HeaderView
  extends HComponent
  implements VectorModelListener,
  AdjustmentListener {

  private int thickness = 0;
  private int orientation =
    Orientation.HORIZONTAL;

  private boolean horz = true;
  private int offset = 0;
  private int origin = 0;
  private VectorModel model;
  private ItemPainter painter =
    new ButtonItemPainter(new TextItemPainter(
    Alignment.CENTER | Alignment.MIDDLE));

  private SizeVector sizes;

  private int selection = -1;
  private int margin = 8;

  private boolean resizing = false;
  private boolean batchMode = false;

  private int clickItem;

  private Image canvas;

  private EventMulticaster headerListeners =
    new EventMulticaster();

  private static Cursor DEFAULT_CURSOR =
    Cursor.getDefaultCursor();

  private static Cursor H_SIZE_CURSOR =
    new Cursor(Cursor.W_RESIZE_CURSOR);

  private static Cursor V_SIZE_CURSOR =
    new Cursor(Cursor.N_RESIZE_CURSOR);

  public HeaderView() {
    this(Orientation.HORIZONTAL);
  }

  public HeaderView(int defaultOrientation) {
    orientation = defaultOrientation;
    horz = (orientation == Orientation.HORIZONTAL);
  }

  public void setModel(VectorModel vm) {
    if (model != null) {
      model.removeModelListener(this);
    }

    model = vm;
    if (model != null) {
      model.addModelListener(this);
    }

    repaint();
  }

  public VectorModel getModel() {
    return model;
  }

  public void setPainter(ItemPainter newPainter) {
    painter = newPainter;
    repaint();
  }

  public ItemPainter getPainter() {
    return painter;
  }

  public void addHeaderListener(HeaderListener listener) {
    headerListeners.add(listener);
  }

  public void removeHeaderListener(HeaderListener listener) {
    headerListeners.remove(listener);
  }

  public int getOrientation() {
    return orientation;
  }

  public void setOrientation(int orientation) {
    this.orientation = orientation;
    horz = (orientation == Orientation.HORIZONTAL);
    invalidate();
    repaintItems();
  }

  public void setBatchMode(boolean batchMode) {
    this.batchMode = batchMode;
    if (!this.batchMode) {
      repaintItems();
    }
  }

  public boolean isBatchMode() {
    return batchMode;
  }

  public void modelStructureChanged(VectorModelEvent e) {
    repaintItems();
  }

  public void modelContentChanged(VectorModelEvent e) {
    repaintItem(e.getLocation());
  }

  public void setOffset(int offset) {
    this.offset = offset;
  }

  public int getOffset() {
    return offset;
  }

  public void setOrigin(int origin) {
    this.origin = Math.max(origin, 0);
    setOffset(sizes.getSizeUpTo(origin));
    repaintItems();
  }

  public int getOrigin() {
    return origin;
  }

  public int getThickness() {
    if (thickness == 0 && model != null) {
      Graphics g = getGraphics();
      if (g == null) {
        return 0;
      }

      Font f = getFont();
      if (horz) {
        g.setFont(f);
        Object data = model.getCount() > 0 ?
          model.get(0) : null;

        Dimension ps = painter != null ?
          painter.getPreferredSize(
          data, g, 0, null) :
          new Dimension(0,0);

        if (model.getCount() > 0) {
          thickness = ps != null ? ps.height : 0;
        } else {
          FontMetrics fm = g.getFontMetrics(f);
          thickness = fm != null ?
            fm.getHeight() : 0;
        }
      } else {
        FontMetrics fm = g.getFontMetrics(f);
        thickness = fm != null ?
          fm.stringWidth("WWW") : 0;
      }
    }

    return thickness;
  }

  public void setThickness(int thickness) {
    if (this.thickness != thickness) {
      this.thickness = thickness;
      getThickness();
      invalidate();
      repaintItems();
    }
  }

  public void setItemSizes(SizeVector newSizes) {
    sizes = newSizes;
  }

  public SizeVector getItemSizes() {
    return sizes;
  }

  public Dimension getPreferredSize() {
    getThickness();
    if (model == null || sizes == null) {
      return new Dimension(thickness, thickness);
    }

    if (horz) {
      return new Dimension(
        sizes.getSizeUpTo(model.getCount()), thickness);
    } else {
      return new Dimension(thickness,
        sizes.getSizeUpTo(model.getCount()));
    }
  }

  private int hitTest(int pos) {
    int item = 0;
    int width = sizes.getSize(item) - offset;
    int count = model.getCount();

    while (pos >= width && ++item < count) {
      width += sizes.getSize(item);
    }

    if (item != 0 && item == count) {
      item--;
    }

    return item;
  }

  protected void processHeaderEvent(HeaderEvent e) {
    headerListeners.dispatch(e);
  }

  protected void processMousePressed(
    MouseEvent e) {

    super.processMousePressed(e);
    int x = e.getX();
    int y = e.getY();
    if (headerListeners.hasListeners()) {
      int item = horz ?
        hitTest(x) : hitTest(y);

      Rectangle rect = getItemBounds(item);
      if (horz) {
        if (x <= rect.x + rect.width) {
          clickItem = item;
          if (x >= (rect.x + rect.width - margin)) {
            resizing = true;
            processHeaderEvent(new HeaderEvent(
              this, HeaderEvent.START_RESIZE,
              clickItem, x, y));
          }
        }
      } else {
        if (y <= rect.y + rect.height) {
          clickItem = item;
          if (y >= (rect.y + rect.height - margin)) {
            resizing = true;
            processHeaderEvent(new HeaderEvent(
              this, HeaderEvent.START_RESIZE,
              clickItem, x, y));
          }
        }
      }
    }
  }

  protected void processMouseMoved(MouseEvent e) {
    super.processMouseMoved(e);
    int pos = horz ? e.getX() : e.getY();
    int item = hitTest(pos);

    Rectangle rect = getItemBounds(item);
    if (horz) {
      if (pos <= rect.x + rect.width) {
        if (pos >= (rect.x + rect.width - margin)) {
          setCursor(H_SIZE_CURSOR);
          return;
        }
      }
    } else {
      if (pos <= rect.y + rect.height) {
        if (pos >= (rect.y + rect.height - margin)) {
          setCursor(V_SIZE_CURSOR);
          return;
        }
      }
    }

    setCursor(DEFAULT_CURSOR);
  }

  protected void processMouseDragged(MouseEvent e) {
    super.processMouseDragged(e);
    if (headerListeners.hasListeners()) {
      if (resizing) {
        int x = e.getX();
        int y = e.getY();
        if (resizing) {
          processHeaderEvent(new HeaderEvent(
            this, HeaderEvent.WHILE_RESIZE,
            clickItem, x, y));

          return;
        }
      }

      setCursor(DEFAULT_CURSOR);
    }
  }

  protected void processMouseReleased(MouseEvent e) {
    super.processMouseReleased(e);
    int x = e.getX();
    int y = e.getY();
    if (headerListeners.hasListeners()) {
      if (resizing) {
        processHeaderEvent(new HeaderEvent(
          this, HeaderEvent.STOP_RESIZE,
          clickItem, x, y));

        repaintItems();
      }
    }

    resizing  = false;
    setCursor(DEFAULT_CURSOR);
  }

  protected void processMouseClicked(MouseEvent e) {
    super.processMouseClicked(e);
    int x = e.getX();
    int y = e.getY();
    int pos = horz ? x : y;
    int index = hitTest(pos);
    if (index == clickItem) {
      processHeaderEvent(new HeaderEvent(
        this, HeaderEvent.ITEM_CLICKED,
        index, x, y));
    }
  }

  public Rectangle getItemBounds(int item) {
    Rectangle rect = getBounds();
    if (rect == null) {
      return new Rectangle();
    }

    rect.x = 0;
    rect.y = 0;

    int l = -offset;
    for (int i = 0; i < item; i++) {
      l += sizes.getSize(i);
    }

    if (horz) {
      rect.x = l;
      rect.width = sizes.getSize(item);
    } else {
      rect.y = l;
      rect.height = sizes.getSize(item);
    }

    return rect;
  }

  public void repaintItem(int index) {
    Rectangle rect = getItemBounds(index);
    Graphics g = getGraphics();
    if (rect != null && g != null) {
      g.setClip(rect.x, rect.y, rect.width, rect.height);
      paint(g);
    }
  }

  public void repaintItems() {
    Graphics g = getGraphics();
    if (g != null) {
      Dimension sz = getSize();
      g.setClip(0, 0, sz.width, sz.height);
      paint(g);
    }
  }

  public void paint(Graphics pg) {
    if (batchMode || painter == null ||
      model == null) {

      return;
    }

    Rectangle clip = getClipBounds();
    if (clip == null) {
      return;
    }

    Dimension vp = getSize();
    if (canvas == null ||
      canvas.getWidth(null) != vp.width ||
      canvas.getHeight(null) != vp.height) {

      canvas = createImage(vp.width, vp.height);
    }

    Graphics g = canvas.getGraphics();
    if (g == null) {
      return;
    }

    g.setColor(getBackground());
    g.fillRect(0, 0, vp.width, vp.height);
    g.translate(-clip.x, -clip.y);
    g.setClip(clip.x, clip.y, clip.width, clip.height);

    super.paint(g);

    int count = model.getCount();
    int vx = horz ? clip.x : clip.y;
    int vw = horz ? clip.width : clip.height;
    int start = Math.min(hitTest(vx), count);
    int end   = Math.min(hitTest(vx + vw), count);

    int viewstate =
      (isEnabled() ? 0 : ItemPainter.DISABLED);

    if (count > 0 && start >= 0) {
      Rectangle rect = getItemBounds(start);
      for (int index = start; index <= end; index++) {
        if (horz) {
          rect.width = sizes.getSize(index);
        } else {
          rect.height = sizes.getSize(index);
        }

        Object value = model.get(index);

        int state = viewstate |
          (index == selection ? ItemPainter.SELECTED : 0);

        g.setColor(getBackground());
        g.setFont(getFont());
        g.fillRect(rect.x, rect.y, rect.width, rect.height);
        g.clipRect(rect.x, rect.y, rect.width, rect.height);
        painter.paint(value, g, rect, state, null);
        g.setClip(clip.x, clip.y, clip.width, clip.height);

        if (horz) {
          rect.x += rect.width;
        } else {
          rect.y += rect.height;
        }
      }
    } else {
      g.setColor(getBackground());
      g.fillRect(clip.x, clip.y, clip.width, clip.height);
    }

    pg.drawImage(canvas, 0, 0, null);
    g.dispose();
  }

  public void adjustmentValueChanged(AdjustmentEvent e) {
    offset = e.getValue();
    repaintItems();
  }
}
