package higa.jhcl.view;

import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;

import java.awt.TextField;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.FontMetrics;
import java.awt.Component;
import java.awt.Toolkit;

import java.text.Format;

import higa.jhcl.model.ItemEditor;
import higa.jhcl.model.ItemEditSite;

import higa.util.Formatter;

public class TextItemEditor extends TextField
  implements ItemEditor, Serializable {

  private transient Format formatter;

  public TextItemEditor() {}

  public TextItemEditor(Format formatter) {
    this.formatter = formatter;
  }

  public void setFormatter(Format formatter) {
    this.formatter = formatter;
  }

  public Format getFormatter() {
    return formatter;
  }

  public void startEdit(Object data,
    Rectangle bounds, ItemEditSite site) {

    String text = Formatter.format(formatter, data);
    setText(text);

    if (site != null) {
      setBackground(site.getBackground());
      setForeground(site.getForeground());
      setFont(site.getFont());
    }

    setBounds(bounds.x, bounds.y,
      bounds.width, bounds.height);

    setVisible(true);

    Point clickPoint = site != null ? site.getEditClickPoint() : null;
    int position = 0;

    if (clickPoint == null && text != null) {
      position = text.length();
      select(0, position);
    } else if (text != null) {
      FontMetrics fm =
        Toolkit.getDefaultToolkit().getFontMetrics(getFont());

      int widths[] = fm.getWidths();
      int i = 0;
      int xClick = clickPoint.x - bounds.x;
      for (; i < text.length(); i++) {
        int width = fm.stringWidth(text.substring(0, i));
        if (width >= xClick) {
          break;
        }
      }

      position = i;
      select(position, position);
    }

    requestFocus();
  }

  public Object getValue() {
    return Formatter.parseObject(formatter, getText());
  }

  public Component getComponent() {
    return this;
  }

  public void endEdit(boolean posted) {}

  private void writeObject(ObjectOutputStream s)
    throws IOException {

    s.defaultWriteObject();
    s.writeObject(
      formatter instanceof Serializable ? formatter : null);
  }

  private void readObject(ObjectInputStream s)
    throws IOException, ClassNotFoundException {

    s.defaultReadObject();
    Object data = s.readObject();
    if (data instanceof Format) {
      formatter = (Format)data;
    }
  }
}