package higa.sql.dataset;

import java.io.Serializable;

public class ReadRow implements Serializable,
  Cloneable {

  protected Object[] items =
    new Object[0];

  private static int _rowid = -1;
  private int rowid = -1;

  public ReadRow() {
    _rowid++;
    rowid = _rowid;
  }

  public ReadRow(Object[] newItems) {
    items = new Object[newItems.length];
    System.arraycopy(newItems, 0,
      items, 0, newItems.length);
  }

  public ReadRow(ReadRow rr) {
    this(rr.items);
    rowid = rr.rowid;
  }

  public Object clone() {
    return new ReadRow(this);
  }

  public int getColumnCount() {
    return items.length;
  }

  public Object get(int columnIndex) {
    if (checkColumn(columnIndex)) {
      Object item = new Object();
      item = items[columnIndex];
      return item;
    } else {
      return null;
    }
  }

  public int getRowid() {
    return rowid;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();
    int count = getColumnCount();
    for (int i = 0; i < count; i++) {
      Object item = get(i);
      String s =  item != null ?
        item.toString() : null;

      if (item instanceof byte[]) {
        s = new String((byte[])item);
      }

      sb.append(s + "\t");
    }

    if (sb.length() > 0) {
      sb.setCharAt(sb.length() - 1, '\n');
    } else {
      sb.append("\n");
    }

    return sb.toString();
  }

  protected boolean checkColumn(int columnIndex) {
    if (columnIndex < 0) {
      throw new IllegalArgumentException(
        "columnIndex:" + columnIndex);
    }

    if (columnIndex < getColumnCount()) {
      return true;
    } else {
      return false;
    }
  }
}