package higa.util;

import java.math.BigDecimal;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.PreparedStatement;
import java.sql.CallableStatement;
import java.sql.Types;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.SQLException;

import java.util.Enumeration;
import java.util.Hashtable;

public class JDBCUtil {

  private static final Resources resources =
    new Resources("higa.util.resources.jdbcutil");
    
  private static final String[] COLUMN = {
    resources.getString("column1"),
    resources.getString("column2"),
    resources.getString("column3")};

  private static final int COLUMN_COUNT = COLUMN.length;

  private static final int[] COLUMN_SIZE = {30, 8, 25};

  public static String getResultSetContents(ResultSet rs) {
    try {
      StringBuffer sb = new StringBuffer();
      ResultSetMetaData rsmd = rs.getMetaData();
      int columnCount = rsmd.getColumnCount();
      int[] displaySizes = new int[columnCount];
      boolean[] numbers = new boolean[columnCount];
      for (int i = 0; i < columnCount; i++) {
        displaySizes[i] = rsmd.getColumnDisplaySize(i + 1);
        numbers[i] = isNumber(rsmd.getColumnType(i + 1));

        String s = getDisplayString(
          rsmd.getColumnName(i + 1), displaySizes[i],
          numbers[i]);

        sb.append(s + " ");
      }

      sb.setCharAt(sb.length() - 1, '\n');

      sb.append(getSeparator(displaySizes));

      while (rs.next()) {
        for (int i = 0; i < columnCount; i++) {
          String s = getDisplayString(
            rs.getString(i + 1), displaySizes[i],
            numbers[i]);

          sb.append(s + " ");
        }

        sb.setCharAt(sb.length() - 1, '\n');
      }

      return sb.toString();
    } catch (Exception ex) {
      ex.printStackTrace();
      return null;
    }
  }

  public static boolean isNumber(int type) {
    boolean b = false;
    switch (type) {
      case Types.BIGINT:
      case Types.DECIMAL:
      case Types.DOUBLE:
      case Types.FLOAT:
      case Types.INTEGER:
      case Types.NUMERIC:
      case Types.REAL:
      case Types.SMALLINT:
      case Types.TINYINT:
        b = true;
        break;
    }

    return b;
  }

  public static boolean isDate(int type) {
    boolean b = false;
    switch (type) {
      case Types.DATE:
      case Types.TIME:
      case Types.TIMESTAMP:
        b = true;
        break;
    }

    return b;
  }

  public static boolean isString(int type) {
    boolean b = false;
    switch (type) {
      case Types.CHAR:
      case Types.VARCHAR:
      case Types.LONGVARCHAR:
        b = true;
        break;
    }

    return b;
  }

  public static String getDisplayString(
    String s, int size) {
    return getDisplayString(s, size, false);
  }

  public static String getDisplayString(
    String s, int size, boolean number) {
    if (s == null) {
      s = " ";
    }

    StringBuffer sb =
      new StringBuffer(s.replace('\n', ' '));

    int blanks = size - s.length();
    blanks = blanks > 0 ? blanks : 0;
    StringBuffer sb2 = new StringBuffer();
    for (int i = 0; i < blanks; i++) {
      sb2.append(" ");
    }

    if (number) {
      sb.insert(0, sb2.toString());
    } else {
      sb.append(sb2.toString());
    }

    sb.setLength(size);
    return sb.toString();
  }

  public static String getSeparator(int[] sizes) {
    StringBuffer sb = new StringBuffer();
    int count = sizes.length;
    for (int i = 0; i < count; i++) {
      for (int j = 0; j < sizes[i]; j++) {
        sb.append("-");
      }
      sb.append(" ");
    }
    sb.setCharAt(sb.length() - 1, '\n');
    return sb.toString();
  }

  public static void setInParameter(
    PreparedStatement stmt, int index,
    Object value, int type) throws SQLException {

    index++;
    if (null == value) {
      stmt.setNull(index, type);
      return;
    }

    switch (type) {
      case Types.BIT:
        stmt.setBoolean(index,
          Boolean.valueOf(value.toString()).booleanValue());

        break;
      case Types.TINYINT:
      case Types.SMALLINT:
      case Types.INTEGER:
        stmt.setInt(index,
          Integer.parseInt(value.toString()));

        break;
      case Types.BIGINT:
        stmt.setLong(index,
          Long.parseLong(value.toString()));

        break;
      case Types.REAL:
      case Types.FLOAT:
      case Types.DOUBLE:
        stmt.setDouble(index,
          Double.valueOf(value.toString()).doubleValue());

        break;
      case Types.NUMERIC:
      case Types.DECIMAL:
        stmt.setBigDecimal(index,
          new BigDecimal(value.toString()));

        break;
      case Types.CHAR:
      case Types.VARCHAR:
      case Types.LONGVARCHAR:
        stmt.setString(index, value.toString());
        break;
      case Types.BINARY:
      case Types.VARBINARY:
      case Types.LONGVARBINARY:
        byte[] bytes = null;
        if (value instanceof byte[]) {
          bytes = (byte[])value;
        } else {
          bytes = value.toString().getBytes();
        }

        stmt.setBytes(index, bytes);
        break;
      case Types.DATE:
        Date date = null;
        if (value instanceof java.util.Date) {
          date = new Date(
            ((java.util.Date)value).getTime());
        } else {
          date = Date.valueOf(value.toString());
        }

        stmt.setDate(index, date);
        break;
      case Types.TIME:
        Time time = null;
        if (value instanceof java.util.Date) {
          time = new Time(
            ((java.util.Date)value).getTime());
        } else {
          time = Time.valueOf(value.toString());
        }

        stmt.setTime(index, time);
        break;
      case Types.TIMESTAMP:
        Timestamp timestamp = null;
        if (value instanceof Timestamp) {
          timestamp = (Timestamp)value;
        } else if (value instanceof java.util.Date) {
          timestamp = new Timestamp(
            ((java.util.Date)value).getTime());
        } else {
          try {
            timestamp = Timestamp.valueOf(value.toString());
          } catch (IllegalArgumentException ex) {
            timestamp = new Timestamp(
              (Date.valueOf(value.toString())).getTime());
          }
        }

        stmt.setTimestamp(index, timestamp);
        break;
      default:
        stmt.setObject(index, value);
        break;
    }
  }

  public static String describe(ResultSetMetaData rsmd)
    throws SQLException {

    StringBuffer sb = new StringBuffer();

    for (int i = 0; i < COLUMN_COUNT; i++) {
      sb.append(getDisplayString(
        COLUMN[i], COLUMN_SIZE[i]) + " ");
    }

    sb.setCharAt(sb.length() - 1, '\n');

    sb.append(getSeparator(COLUMN_SIZE));

    int count = rsmd.getColumnCount();
    for (int i = 1; i <= count; i++) {
      String columnName = rsmd.getColumnName(i);
      sb.append(getDisplayString(
        columnName, COLUMN_SIZE[0]) + " ");

      String nullable =
        rsmd.isNullable(i) == 0 ? "NOT NULL" : "";
      sb.append(getDisplayString(
        nullable, COLUMN_SIZE[1]) + " ");

      StringBuffer dataType = new StringBuffer(
        rsmd.getColumnTypeName(i));

      int precision = rsmd.getPrecision(i);
      if (precision > 0) {
        dataType.append("(" + String.valueOf(precision));
      }

      int scale = rsmd.getScale(i);
      if (scale > 0) {
        dataType.append("," + String.valueOf(scale));
      }

      if (precision > 0) {
        dataType.append(")");
      }

      sb.append(getDisplayString(
        dataType.toString(), COLUMN_SIZE[2]) + "\n");
    }

    return sb.toString();
  }

  public static ResultSet execute(Connection con, String sql)
    throws SQLException {

    return execute(con, sql, null, null);
  }

  public static ResultSet execute(Connection con, String sql,
    Hashtable inParameter, Hashtable outParameter)
    throws SQLException {

    CallableStatement cstmt = con.prepareCall(sql);

    if (inParameter != null && !inParameter.isEmpty()) {
      Enumeration keys = inParameter.keys();
      Enumeration elements = inParameter.elements();
      while (keys.hasMoreElements()) {
        Integer integer = (Integer)keys.nextElement();
        int index = integer.intValue();
        String value = (String)elements.nextElement();
        cstmt.setString(index, value);
      }
    }

    if (outParameter != null && !outParameter.isEmpty()) {
      Enumeration keys = outParameter.keys();
      while (keys.hasMoreElements()) {
        Integer integer = (Integer)keys.nextElement();
        int index = integer.intValue();
        cstmt.registerOutParameter(
          index, Types.VARCHAR);
      }
    }

    if (cstmt.execute(sql)) {
      return cstmt.getResultSet();
    } else {
      if (outParameter != null && !outParameter.isEmpty()) {
        Enumeration keys = outParameter.keys();
        while (keys.hasMoreElements()) {
          Integer integer = (Integer)keys.nextElement();
          int index = integer.intValue();
          String value = cstmt.getString(index);
          outParameter.remove(integer);
          outParameter.put(integer, value);
        }
      }

      return null;
    }
  }
}